import time
import itertools
import numpy as np


def run_loop(duration=None, rate=None, interval=None, n=None, delay=1e-5):
    loop_ = sleep_loop(delay) if delay else loop()
    return throttled(
        timed(limit(loop_, n), duration=duration),
        rate, interval, delay=delay or 0)

def loop(i=0):
    '''Infinite loop'''
    while True:
        yield i
        i += 1

def sleep_loop(delay=1e-5):
    for _ in loop():
        yield _
        time.sleep(delay)

def timed(it=None, duration=None, error=False):
    '''Run a loop for a predetermined amount of time.'''
    if isinstance(it, (float, int)):
        duration, it = it, loop()
    if it is None:
        it = loop()
    if not duration:
        yield from it
        return

    t0 = time.time()
    for x in it:
        if time.time() - t0 >= duration:
            if error:
                raise TimeoutError('Loop did not exit in under {} seconds.'.format(duration))
            break
        yield x


def throttled(it=None, rate=None, interval=None, delay=1e-5, initial=None):
    '''Throttle a loop to take '''
    if initial:
        time.sleep(initial)
    if it is None:
        it = loop()
    rate = 1. / interval if interval else rate
    if not rate:
        yield from it
        return

    t1 = time.time()
    for x in it:
        yield x
        t1, t0 = time.time(), t1
        dt = max((1. / rate) - (t1 - t0 + 0.5 * delay), 0)
        if dt:
            time.sleep(dt)
        # wraps around for _ in because that may take a non-trivial amount of time
        # e.g. (time.sleep(1) for _ in loop())
        t1 = time.time()


def thread_sleep(it, delay=1e-5):
    if not delay:
        yield from it
        return

    for _ in it:
        yield _
        time.sleep(delay)


def resample_iter(it, interval):
    if not interval:
        yield from it
        return

    t = time.time()
    for _ in it:
        if time.time() - t >= interval:
            t = time.time()
            yield _


def limit(it, n=None):
    if not n:
        yield from it
        return
    # yield only n elements
    for i, x in zip(range(n), it):
        yield x


def peakiter(it, n=1):
    '''Check the value first n items of an iterator without unloading them from
    the iterator queue.'''
    it = iter(it)
    items = [_ for _, i in zip(it, range(n))]
    return items, itertools.chain(items, it)


# numpy

def npgenarray(it, shape, **kw):
    '''Create a np.ndarray from a generator. Must specify at least the length
    of the generator or the entire shape of the final array.'''
    if isinstance(shape, int):
        (x0,), it = peakiter(it)
        shape = (shape,) + x0.shape
    X = np.empty(shape, **kw)
    for i, x in enumerate(it):
        X[i] = x
    return X


def npmaparray(func, X):
    return npgenarray((func(x) for x in X), len(X))


# if __name__ == '__main__':
# for _ in timed(throttled(loop(), 1), 10):
#     ...



import os
import soundfile
import reip
# import time


class AudioFile(reip.Block):
    def __init__(self, filename='{time}.wav'):
        self.filename = filename
        super().__init__()

    def process(self, X, meta):
        # meta["time"] = time.time()
        fname = self.filename.format(**meta)
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        soundfile.write(fname, X, meta['sr'])
        return [fname], meta



import time
import re
import queue
import sounddevice as sd
import numpy as np
import reip


class Mic(reip.Block):
    _audio_stream = device = blocksize = None
    def __init__(self, device=None, sr=None, block_duration=1, channels=None,
                 mono=False, search_interval=10, dtype=np.int16, mic_compensation=None, skip_n=0, **kw):
        self.device_name = device
        self.sr = sr
        self.channels = channels
        self.mono = (0 if mono is True else mono if mono is not False else None)
        self.dtype = np.dtype(dtype)
        self._is_float = self.dtype == np.float32
        self.block_duration = block_duration
        self.search_interval = search_interval
        self.mic_compensation = mic_compensation
        if self.mic_compensation is not None:
            from scipy.signal import lfilter, lfilter_zi
            self._lfilter = lfilter  # only import once
            self.filter_zi = lfilter_zi(self.mic_compensation, [1.0])
        self.skip_n = skip_n or 0
        super().__init__(**kw)

        self._q = reip.stores.Producer()
        self.sources[0] = self._q.gen_source()

    # def init(self):
    #     '''Start pyaudio and start recording'''
    #     for i in reip.util.iters.run_loop(interval=self.search_interval):
    #         try:
    #             self._init()
    #             break
    #         except Exception as e:
    #             self.log.error('Microphone Init: ({}) {}'.format(e.__class__.__name__, e))
    #             if i == 0:
    #                 self.log.exception(e)

    def init(self):
        self.n_frames = 0
        # initialize pyaudio
        device = find_device(self.device_name)
        self.log.info('Using audio device: {} - {}'.format(device['name'], device))

        # get parameters from device
        self.device = device
        self.sr = int(self.sr or device['default_samplerate'])
        self.channels = self.channels or device['max_input_channels']
        self.blocksize = int(self.block_duration * self.sr)

        # start audio streamer
        self._audio_stream = sd.InputStream(
            device=self.device.get('index', self.device['name']), blocksize=self.blocksize,
            samplerate=self.sr, channels=self.channels, dtype=self.dtype,
            callback=self._stream_callback)
        self._audio_stream.start()

    def _stream_callback(self, buf, frames, time_info, status):
        '''Append frames to the queue - blocking API is suuuuuper slow.'''
        timestamp = time.time() - 1.*len(buf)/self.sr
        if status:
            self.log.error('Input overflow status: {}'.format(status))

        self.n_frames += 1
        if self.n_frames <= self.skip_n:
            return
        try:
            with self._except('process'):
                self._q.put((np.copy(buf), {'time': timestamp}))
        except Exception as e:
            self.log.exception(e)

    def process(self, pcm, meta):
        if not self._is_float:
            pcm = pcm / float(np.iinfo(pcm.dtype).max)
        if self.mono is not None:
            pcm = np.ascontiguousarray(pcm[:,self.mono])
        if self.mic_compensation is not None:
            pcm, self.filter_zi = self._lfilter(self.mic_compensation, [1.0], pcm, zi=self.filter_zi)
        return [pcm], {'sr': self.sr, 'time': meta["time"]}

    def finish(self):
        '''Stop pyaudio'''
        if self._audio_stream is not None:
            self._audio_stream.stop()
            self._audio_stream.close()
        self._audio_stream = None
        self.device = None


def find_device(query, min_input=1, min_output=0, log=None):
    '''Search for an audio device by name.'''
    devices = sd.query_devices(query or sd.default.device, kind='input')
    if devices:
        return devices
    devices = [dict(d, index=i) for i, d in enumerate(sd.query_devices())]
    if log is not None:
        log.debug('Available Audio Devices: {}'.format(devices))
    try:
        return next((
            d for d in devices
            if d['max_input_channels'] >= min_input
            and d['max_output_channels'] >= min_output
            and (query is None or re.search(query, d['name'], re.IGNORECASE))
        ))
    except StopIteration:
        raise OSError('No device found matching "{}" in {}.'.format(
            query, [d['name'] for d in devices]))



import reip
import numpy as np
import tflit
# import reip_gui.builder as rgui


class Tflite(reip.Block):
    '''Run a tflite model.'''
    model = None
    # @rgui.options(rgui.Text(), rgui.MultiLabel(), input_features=rgui.FunctionDef('data', 'meta'))
    # @rgui.options  # infer from docstring
    def __init__(self, filename, labels=None, input_features=None, **kw):
        '''Run a tflite model on the data input.

        Arguments:
            filename (str): the model filename.
            labels (list[str]): the labels corresponding to each model output.
            input_features (callable): prepare the block input so that it is in
                the format needed by the model.
        '''
        super().__init__(**kw)
        self.filename = filename
        self.labels = labels
        self._get_input_features = (
            self.input_features if input_features is None
            else input_features)

    def input_features(self, data, meta):
        return data

    def init(self):
        self.model = tflit.Model(self.filename)

    def process(self, X, meta):
        meta["labels"] = self.labels
        return (
            [self.model.predict(self._get_input_features(X, meta))],
            meta
        )
        # return (
        #     [self.model.predict(self._get_input_features(X, meta))],
        #     {'labels': self.labels}
        # )

    def finish(self):
        self.model = None


# def load_tflite_model_function(model_path):
#     import tflite_runtime.interpreter as tflite
#     compute = prepare_model_function(tflite.Interpreter(model_path))
#     compute.model_path = model_path
#     return compute
#
#
# def prepare_model_function(model, verbose=False):
#     # allocate and get shapes
#     in0_dets = model.get_input_details()[0]
#     out0_dets = model.get_output_details()[0]
#     input_shape, output_shape = in0_dets['shape'][1:], out0_dets['shape'][1:]
#     in_idx, out_idx = in0_dets['index'], out0_dets['index']
#     model.allocate_tensors()
#
#     if verbose:
#         print('-- Input details --')
#         print(in0_dets, '\n')
#         print('-- Output details --')
#         print(out0_dets, '\n')
#
#     # Get the L3 embedding
#     def compute(x):
#         model.set_tensor(in_idx, x.copy().astype(np.float32))
#         model.invoke()
#         return model.get_tensor(out_idx).copy()
#     compute.input_shape = input_shape
#     compute.output_shape = output_shape
#     return compute



import os
import cv2
import reip
import tflit
import numpy as np

def _model_file(name):
    return os.path.join(os.path.dirname(reip.__file__), 'models', name)

class COCOMobileNet(reip.Block):
    '''
    Object detection model from tensorflow. 80 classes.

    https://github.com/EdjeElectronics/TensorFlow-Lite-Object-Detection-on-Android-and-Raspberry-Pi/blob/master/TFLite_detection_webcam.py

    Outputs:
      - boxes - Bounding box coordinates of detected objects
      - classes - Class index of detected objects
      - scores - Confidence of detected objects
      - num - Total number of detected objects (inaccurate and not needed)

    '''
    MODEL_FILE = "coco_ssd_mobilenet_v1_1.0_quant_2018_06_29.tflite"
    LABEL_FILE = "coco_ssd_mobilenet_v1_1.0_quant_2018_06_29_labelmap.txt"

    def __init__(self, threshold=0.5, draw=None, **kw):
        self._should_draw = draw
        self.threshold = threshold
        super().__init__(n_outputs=2, **kw)

    def init(self):
        self.model = tflit.Model(reip.package_file('models', self.MODEL_FILE))
        # print(self.model.input_shape, self.model.output_shape, self.model.dtype)
        self.size = self.model.input_shape[1:-1]

        with open(reip.package_file('models', self.LABEL_FILE), 'r') as f:
            self.labels = [line.strip() for line in f.readlines()]
        if self.labels[0] == '???':
            del self.labels[0]

        self.should_draw = bool(self.sinks[1].readers) if self._should_draw is None else self._should_draw
        self.log.debug(f'will draw: {self.should_draw}')

    def process(self, frame, meta):
        X = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        X = cv2.resize(X, self.size)
        X = (np.float32(X) - 127.5) / 127.5
        boxes, classes, scores, _ = self.model.predict(X[None, ...])
        return [
            [boxes, classes, scores],
            draw_boxes(frame.copy(), boxes[0], classes[0], scores[0], self.labels, self.threshold)
            if self.should_draw else None
        ], meta


def draw_boxes(img, boxes, classes, scores, labels, threshold=0.5):
    imH, imW = img.shape[:2]
    for i in range(len(scores)):
        if threshold < scores[i] <= 1.0:
            # Get bounding box coordinates and draw box
            # Interpreter can return coordinates that are outside of image dimensions,
            # need to force them to be within image using max() and min()
            ymin = int(max(1, (boxes[i][0] * imH)))
            xmin = int(max(1, (boxes[i][1] * imW)))
            ymax = int(min(imH, (boxes[i][2] * imH)))
            xmax = int(min(imW, (boxes[i][3] * imW)))
            cv2.rectangle(img, (xmin, ymin), (xmax, ymax), (10, 255, 0), 2)

            # Draw label
            label = '{}: {:.2%}'.format(labels[int(classes[i])], scores[i])
            labelSize, baseLine = cv2.getTextSize(
                label, cv2.FONT_HERSHEY_SIMPLEX, 0.7, 2) # Get font size
            # Make sure not to draw label too close to top of window
            label_ymin = max(ymin, labelSize[1] + 10)
            cv2.rectangle(
                img, (xmin, label_ymin-labelSize[1]-10),
                (xmin+labelSize[0], label_ymin+baseLine-10),
                (10, 255, 0), cv2.FILLED) # Draw white box to put label text in
            cv2.putText(
                img, label, (xmin, label_ymin-7),
                cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0,0,0), 2) # Draw label text

    # # Draw framerate in corner of frame
    # cv2.putText(img, 'FPS: {0:.2f}'.format(frame_rate_calc), (30,50),
    #             cv2.FONT_HERSHEY_SIMPLEX, 1, (255, 255, 0), 2, cv2.LINE_AA)
    return img


Objects = COCOMobileNet


from collections import ChainMap


class Meta(ChainMap):
    '''Merged metadata dict stacks into one. Removes duplicates and empty maps.
    The original input maps can be accessed using `[d for d in stack.sources]`
    '''
    def __init__(self, *maps, filter=False):
        self.sources = maps
        super().__init__(*ordered_unique(flatten_maps(*maps, filter=filter)))

    def update(self, *a, **kw):
        super().update(*a, **kw)
        return self

    def set_defaults(self, **kw):
        for k, v in kw.items():
            self.setdefault(k, v)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self.sources[key]
        return super().__getitem__(key)


def flatten_maps(*metas, filter=False):
    '''Flatten nested chain maps.'''
    for m in metas:
        if isinstance(m, ChainMap):
            yield from flatten_maps(*m.maps, filter=filter)
        elif not filter or m:
            yield m


def ordered_unique(maps):
    '''Remove duplicate maps (may happen due to branching+merging).'''
    ids = set()
    for m in maps:
        i = repr(m)
        if i not in ids:
            ids.add(i)
            yield m



import time
import reip


def skip_strategy(source):
    while not source.empty() and source._skip_id < source.skip:
        # source._get()
        source.next()  # discard the buffer
        source._skip_id += 1
        source.skipped += 1

    if source._skip_id == source.skip and not source.empty():
        buffer = source._get()
        source._skip_id = 0  # might not process this buffer in multi_source block
        return buffer


def latest_strategy(source):
    while not source.last():
        # source._get()
        source.next()  # discard the buffer
        source.skipped += 1
    return source._get()


def all_strategy(source):
    if not source.empty():
        return source._get()


class Sink:
    '''An abstract interface representing a place to put data.
    
    Essentially, a put-only queue.
    '''
    def __init__(self):
        self.dropped = 0
        self._full_delay = 1e-6

    def spawn(self):
        '''An optional method that allows a sink to perform initialization.'''
        pass

    def join(self):
        '''An optional method that allows a sink to perform cleanup.'''
        pass

    def __len__(self):
        '''How many elements are in the queue.'''
        raise NotImplementedError

    def full(self):
        '''Is the queue full?'''
        raise NotImplementedError

    def _put(self, buffer):
        '''The overrideable method to put items in the queue.'''
        raise NotImplementedError

    def wait(self, timeout=None):
        '''Wait until the queue has space for another element.'''
        t0 = time.time()
        while self.full():
            time.sleep(self._full_delay)
            if timeout and time.time() - t0 > timeout:
                return None  # QUESTION: TimeoutError instead?

    def put(self, buffer, block=True, timeout=None):
        '''Put an element in the sink.'''
        if block:
            self.wait(timeout)
        return self.put_nowait(buffer)

    def put_nowait(self, buffer):
        '''Put an element in the sink, dropping the value if the queue is full.'''
        if self.full():
            self.dropped += 1
        else:
            self._put(buffer)

    def gen_source(self, **kw):
        raise NotImplementedError


class Source:
    '''An abstract interface representing a place to get data.
    
    Essentially, a get-only queue.
    '''
    # TODO: make it easier for someone to register a new strategy
    #       e.g.:
    #         @Source.register
    #         def SkipStochastic(source): ...
    All = 'all'
    Latest = 'latest'
    Skip = 'skip'

    strategies = {
        All: all_strategy,
        Latest: latest_strategy,
        Skip: skip_strategy,
    }

    _strategy = All

    def __init__(self, strategy=All, skip=0, default=None):
        self.strategy = strategy
        self.skip = skip
        self._skip_id = 0
        self.skipped = 0
        self._empty_delay = 1e-6
        self._default = reip.util.as_func(default) if default is not None else default

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, strategy):
        if strategy not in self.strategies:
            raise ValueError("Unknown strategy '{}'".format(strategy))

        self._strategy = strategy
        self._strategy_get = self.strategies[strategy]

    def __len__(self):
        '''How many elements are in the Source queue.'''
        raise NotImplementedError

    def empty(self):
        '''Is the Source queue empty?'''
        return not len(self) and self._default is None

    def last(self):
        '''Returns True if there is only one element left.'''
        ln = len(self)
        return ln == 1 or (not ln and self._default is not None)

    def next(self):
        raise NotImplementedError

    def _get(self):
        raise NotImplementedError

    def wait(self, timeout=None):
        '''Wait until an element is ready.'''
        t0 = time.time()
        while self.empty():
            time.sleep(self._empty_delay)
            if timeout and time.time() - t0 > timeout:
                return None  # QUESTION: TimeoutError instead?

    def get(self, block=True, timeout=None):
        '''Get the next element in the queue.'''
        if block:
            self.wait(timeout)
        return self.get_nowait()

    def get_nowait(self):
        '''Get the next element in the queue, returning None if no value is available.'''
        if len(self) == 0:
            if self._default is not None:
                return self._default()
            return None
        return self._strategy_get(self)



import time
import threading
import warnings
from contextlib import contextmanager
import remoteobj

import reip
from reip.stores import Producer
from reip.util import text, Meta

'''

'''

__all__ = ['Block']


class _BlockSinkView:
    '''This is so that we can select a subview of a block's sinks. This is similar
    in principal to numpy views. The alternative is to store the state on the original
    object, but that would have unintended consequences if someone tried to create two
    different views from the same block.
    '''
    def __init__(self, block, idx):
        self._block = block
        self._index = idx

    @property
    def item(self):  # can be a single sink, or multiple i.e. block[0] or block[:2]
        return self._block.sinks[self._index]

    @property
    def items(self):  # always a list
        return reip.util.as_list(self.item)

    def __iter__(self):  # iterate over sink
        return iter(self.items)

    def __getitem__(self, key):  # get sub index
        return self.items[key]

    def to(self, *others, squeeze=True, **kw):
        '''Connect this blocks sinks to other blocks' sources.
        
        .. code-block:: python

            InputBlock().to(ProcessBlock())
        '''
        outs = [other(self.items, **kw) for other in others]
        return outs[0] if squeeze and len(outs) == 1 else outs

    def output_stream(self, strategy='all', source_strategy=all, **kw):
        '''Create a Stream iterator which will let you iterate over the
        outputs of a block.'''
        return reip.Stream([s.gen_source(strategy=strategy, **kw) for s in self.items], strategy=source_strategy, **kw)


class Block:
    '''This is the base instance of a block.

    Arguments:
        queue (int): the maximum queue length.
        n_inputs (int, or None): the number of sources to expect. If None,
            it will accept a variable number of sources.
        n_outputs (int): the number of sinks to expect.
             - If `n_outputs=0`, then there will be a single sink, that outputs no
               buffer value, but does output metadata.
               e.g.
               ```python
               # block 1
               return {'just some metadata': True}
               # or
               return [], {'just some metadata': True}

               # block 2
               def process(self, meta):  # no buffer value
                   assert meta['just some metadata'] == True
               ```
             - If `n_outputs=None`, then the block will have no sinks.

        blocking (bool): should the block wait for sinks to have a free space
            before processing more items?
        max_rate (int): the maximum number of iterations per second. If the block
            processes items any faster, it will throttle and sleep the required
            amount of time to meet that requirement.
            e.g. if max_rate=5, then an iteration will take at minimum 0.2s.
        graph (reip.Graph, reip.Task, or False): the graph instance to be added to.
        name ()
    '''
    _thread = None
    _stream = None
    _delay = 1e-4
    parent_id, task_id = None, None
    started = ready = done = closed = _terminated = False
    processed = 0
    controlling = False
    max_rate = None

    def __init__(self, n_inputs=1, n_outputs=1, queue=100, blocking=False, print_summary=True,
                 max_rate=None, min_interval=None, max_processed=None, graph=None, name=None,
                 source_strategy=all, extra_kw=True, extra_meta=None, log_level=None,
                 handlers=None, modifiers=None, input_modifiers=None, **kw):
        self._except = remoteobj.LocalExcept(raises=True)
        self.name = name or f'{self.__class__.__name__}_{id(self)}'
        self.parent_id, self.task_id = reip.Graph.register_instance(self, graph)
        self._print_summary = print_summary

        # sources and sinks
        # by default, a block takes a variable number of sources.
        # If a block has a fixed number of sources, you can specify that
        # and it will throw an exception when spawning if the number of
        # sources does not match.
        self._block_sink_queue_length = queue  # needed for set sink count
        self.n_expected_sources = n_inputs
        self.sources, self.sinks = [], []
        self.set_block_source_count(n_inputs)
        self.set_block_sink_count(n_outputs)
        # used in Stream class. Can be all(), any() e.g. source_strategy=all
        self._source_strategy = source_strategy
        # handlers are wrapper functions that can do things like catch errors and restart when a block is closing.
        self.handlers = reip.util.as_list(handlers or [])
        # modifiers are functions that can be used to alter the output of the block before they are
        # sent to a sink
        self.modifiers = reip.util.as_list(modifiers or [])
        self.input_modifiers = reip.util.as_list(input_modifiers or [])

        if min_interval and max_rate:
            warnings.warn((
                'Both max_rate ({}) and min_interval ({}) are set, but are '
                'mutually exclusive (max_rate=1/min_interval). min_interval will'
                'be used.').format(max_rate, min_interval))
        if max_rate:
            self.max_rate = max_rate
        if min_interval:
            self.min_interval = min_interval
        self.max_processed = max_processed
        self._put_blocking = blocking

        self._extra_meta = reip.util.as_list(extra_meta or [])
        if self._extra_meta:
            # this will flatten a list of dicts into a single dict. Since call=False,
            # if there is a list of dictionaries and a function, any dictionaries
            # before the function will be collapsed, and any after will be left.
            # Calling again (without call=False), will evaluate fully and return
            # the a flattened dict.
            self._extra_meta = reip.util.mergedict(self._extra_meta, call=False)

        if extra_kw:
            self.extra_kw = kw
            for key, value in kw.items():
                setattr(self, key, value)
        elif kw:
            raise TypeError('{} received {} unexpected keyword arguments: {}.'.format(
                self, len(kw), set(kw)))

        # block timer
        self._sw = reip.util.Stopwatch(self.name)
        self.log = reip.util.logging.getLogger(self, level=log_level)
        # signals
        self._reset_state()

    @property
    def min_interval(self):
        return self.max_rate

    @min_interval.setter
    def min_interval(self, value):
        self.max_rate = 1. / value if value is not None else value

    def set_block_source_count(self, n):
        # NOTE: if n is -1, no resize will happen
        self.sources = reip.util.resize_list(self.sources, n, None)

    def set_block_sink_count(self, n):
        # NOTE: if n is None, no resize will happen
        new_sink = lambda: Producer(self._block_sink_queue_length, task_id=self.task_id)
        self.sinks = reip.util.resize_list(self.sinks, n, new_sink)

    def _reset_state(self):
        # state
        self.started = False
        self.ready = False
        self.done = False
        self.closed = False
        self._terminated = False
        # stats
        self.processed = 0
        self.old_processed = 0
        self.old_time = time.time()
        self._sw.reset()
        self._except.clear()

    def __repr__(self):
        return 'Block({}): ({}/{} in, {} out; {} processed) - {}'.format(
            self.name, sum(s is not None for s in self.sources),
            self.n_expected_sources,
            len(self.sinks), self.processed,
            self.block_state_name)

    @property
    def block_state_name(self):
        return (
            (type(self._exception).__name__
             if self._exception is not None else 'error') if self.error else
            ('terminated' if self.terminated else 'done') if self.done else
            ('running' if self.running else 'paused') if self.ready else
            'started' if self.started else '--'  # ??
        )  # add "uptime 35.4s"

    # Graph definition

    def __call__(self, *others, index=0, **kw):
        '''Connect other block sinks to this blocks sources.
        If the blocks have multiple sinks, they will be passed as additional
        inputs.

        .. code-block:: python

            ProcessBlock()(InputBlock())
        '''
        j = next(
            (i for i, s in enumerate(self.sources) if s is None), len(self.sources)
        ) if index == -1 else index
        for i, other in enumerate(others):
            # permit argument to be a block
            # permit argument to be a sink or a list of sinks
            if isinstance(other, Block):
                sinks = other.sinks
            elif isinstance(other, _BlockSinkView):
                sinks = other.items
            else:
                sinks = reip.util.as_list(other)

            if sinks:
                # make sure the list is long enough
                self.set_block_source_count(j + len(sinks))

                # create and add the source
                for j, sink in enumerate(sinks, j):
                    self.sources[j] = sink.gen_source(
                        task_id=self.task_id, **kw)
                j += 1  # need to increment cursor so we don't repeat last index
        return self

    def to(self, *others, squeeze=True, **kw):
        '''Connect this blocks sinks to other blocks' sources.
        
        .. code-block:: python

            InputBlock().to(ProcessBlock())
        '''
        outs = [other(self, **kw) for other in others]
        return outs[0] if squeeze and len(outs) == 1 else outs

    def __or__(self, other):
        '''Connect blocks using Unix pipe syntax.
        
        .. code-block:: python

            InputBlock() | ProcessBlock()
        '''
        return self.to(*reip.util.as_list(other))

    def __ror__(self, other):
        '''Connect blocks using Unix pipe syntax. See :meth:`__or__`'''
        return self(*reip.util.as_list(other))

    def output_stream(self, **kw):
        '''Create a Stream iterator which will let you iterate over the
        outputs of a block.'''
        return reip.Stream.from_block(self, **kw)

    def __getitem__(self, key):
        return _BlockSinkView(self, key)

    # User Interface

    def init(self):
        '''Initialize the block.'''

    def process(self, *xs, meta=None):
        '''Process data.'''
        return xs, meta

    def finish(self):
        '''Cleanup.'''

    # main process loop

    # TODO common worker class
    def run(self, duration=None, **kw):
        with self.run_scope(**kw):
            self.wait(duration)

    @contextmanager
    def run_scope(self, raise_exc=True):
        try:
            self.spawn()
            yield self
        except KeyboardInterrupt:
            self.terminate()
        finally:
            self.join(raise_exc=raise_exc)

    def wait(self, duration=None):
        for _ in reip.util.iters.timed(reip.util.iters.sleep_loop(self._delay), duration):
            if self.done or self.error:
                return True

    def _main(self, _ready_flag=None, duration=None):
        '''The main thread target function. Handles uncaught exceptions and
        generic Block context management.'''
        try:
            self.started = True
            self.closed = False
            self.log.debug(text.blue('Starting...'))
            time.sleep(self._delay)

            with self._sw(), self._except(raises=False):
                # create a stream from sources with a custom control loop
                self._stream = reip.Stream.from_block_sources(
                    self, name=self.name, _sw=self._sw,
                    strategy=self._source_strategy)

                # self.__first_time = True
                # this lets us wrap the block's run function with retry loops, error suppression and
                # whatever else might be useful
                run = reip.util.partial(self.__main_run, _ready_flag=_ready_flag, duration=duration)
                for wrapper in self.handlers[::-1]:
                    run = reip.util.partial(wrapper, self, run)
                run()
        finally:
            try:
                # propagate stream signals to sinks e.g. CLOSE
                if self._stream.signal is not None:
                    self._send_sink_signal(self._stream.signal)
            finally:
                self.done = True
                self.started = False
                self.closed = True
                self.log.debug(text.green('Done.'))


    def __main_run(self, _ready_flag=None, duration=None):
        # # This is to return from a retry loop that doesn't want to close
        # if not self.__first_time and self.closed:
        #     return
        # self.__first_time = False
        try:
            with self._stream:
                self._sw.tick()  # offset time delay due to the reip initialization (e.g. plasma store warm-up)
                # block initialization
                with self._sw('init'): #, self._except('init')
                    self.init()
                self.ready = True
                self.log.debug(text.green('Ready.'))

                if _ready_flag is not None:
                    with self._sw('wait'):
                        _ready_flag.wait()

                # the loop
                loop = reip.util.iters.throttled(
                    reip.util.iters.timed(reip.util.iters.loop(), duration),
                    self.max_rate, delay=self._delay)

                self.old_time = time.time()  # offset self.init() time delay for proper immediate speed calculation

                for _ in self._sw.iter(loop, 'sleep'):
                    inputs = self._stream.get()
                    if inputs is None:
                        break

                    # process each input batch
                    with self._sw('process'): #, self._except('process')
                        buffers, meta = inputs

                        if self._extra_meta:
                            meta.maps += reip.util.flatten(self._extra_meta, call=True, meta=meta)
                        for func in self.input_modifiers:
                            buffers, meta = func(*buffers, meta=meta)
                        outputs = self.process(*buffers, meta=meta)
                        for func in self.modifiers:
                            outputs = func(outputs)

                    # This block of code needs to be here or else self.processed is not counting calls to self.process() function
                    # but buffers generated by current block and thus self.processed will be zero or inacurate in a number of cases:
                    # (i) black hole block that is only consuming data, (ii) data source block that has a buitin buffer bundling/grouping capabilities
                    # We can always add another self.generated counter if we need/want to
                    self.processed += 1
                    # limit the number of blocks
                    if self.max_processed and self.processed >= self.max_processed:
                        self.close(propagate=True)

                    # send each output batch to the sinks
                    with self._sw('sink'):
                        self.__send_to_sinks(outputs, meta)

        except KeyboardInterrupt as e:
            self.log.info(text.yellow('Interrupting'))
            self.log.exception(e)
            # reip.util.print_stack('Interrupted here')
        except Exception as e:
            self.log.exception(e)
            raise
        finally:
            self.ready = False
            # finish up and shut down
            self.log.debug(text.yellow('Finishing...'))
            with self._sw('finish'): # , self._except('finish', raises=False)
                self.finish()

    def __send_to_sinks(self, outputs, meta_in=None):
        '''Send the outputs to the sink.'''
        source_signals = [None]*len(self.sources)
        # retry all sources
        for outs in outputs if reip.util.is_iter(outputs) else (outputs,):
            if outs == reip.RETRY:
                source_signals = [reip.RETRY]*len(self.sources)
            elif outs == reip.CLOSE:
                self.close(propagate=True)
            elif outs == reip.TERMINATE:
                self.terminate(propagate=True)
            # increment sources but don't have any outputs to send
            elif outs is None:
                pass
            #     self._stream.next()
            # increment sources and send outputs
            else:
                # See self.__main_run()
                # self.processed += 1
                # # limit the number of blocks
                # if self.max_processed and self.processed >= self.max_processed:
                #     self.close(propagate=True)

                # detect signals meant for the source
                if self.sources:
                    if outs is not None and any(any(t.check(o) for t in reip.SOURCE_TOKENS) for o in outs):
                        # check signal values
                        if len(outputs) > len(self.sources):
                            raise RuntimeError(
                                'Too many signals for sources in {}. Got {}, expected a maximum of {}.'.format(
                                    self, len(outputs), len(self.sources)))
                        for i, o in enumerate(outs):
                            if o is not None:
                                source_signals[i] = o
                        continue

                # convert outputs to a consistent format
                outs, meta = prepare_output(outs, input_meta=meta_in)
                # pass to sinks
                for sink, out in zip(self.sinks, outs):
                    if sink is not None:
                        sink.put((out, meta), self._put_blocking)

                # increment sources
                # self._stream.next()
        for src, sig in zip(self.sources, source_signals):
            if sig is reip.RETRY:
                pass
            else:
                src.next()


    def _send_sink_signal(self, signal, block=True, meta=None):
        '''Emit a signal to all sinks.'''
        self.log.debug(text.yellow(text.l_('sending', signal)))
        for sink in self.sinks:
            if sink is not None:
                sink.put((signal, meta or {}), block=block)

    # Thread management

    def spawn(self, wait=True, _controlling=True, _ready_flag=None):
        '''Spawn the block thread'''
        try:
            self.controlling = _controlling
            self.log.debug(text.blue('Spawning...'))
            # print(self.summary())

            self._check_source_connections()
            # spawn any sinks that need it
            for s in self.sinks:
                if hasattr(s, 'spawn'):
                    s.spawn()
            self._reset_state()
            self.resume()
            self._thread = remoteobj.util.thread(self._main, _ready_flag=_ready_flag, daemon_=True, raises_=False)
            # threading.Thread(target=self._main, kwargs={'_ready_flag': _ready_flag}, daemon=True)
            self._thread.start()

            if wait:
                self.wait_until_ready()
            if self.controlling:
                self.raise_exception()
        finally:
            # thread didn't start ??
            if self._thread is None or not self._thread.is_alive():
                self.done = True

    def _check_source_connections(self):
        '''Check if there are too many sources for this block.'''
        # check for any empty sources
        disconnected = [i for i, s in enumerate(self.sources) if s is None]
        if disconnected:
            raise RuntimeError(f"Sources {disconnected} in {self} not connected.")

        # check for exact source count
        if (self.n_expected_sources is not None and self.n_expected_sources != -1
                and len(self.sources) != self.n_expected_sources):
            raise RuntimeError(
                f'Expected {self.n_expected_sources} sources '
                f'in {self}. Found {len(self.sources)}.')

    def remove_extra_sources(self, n=None):
        n = n or self.n_expected_sources
        if n is not None and n != -1:
            self.sources = self.sources[:n]

    def wait_until_ready(self):
        '''Wait until the block is initialized.'''
        while not self.ready and not self.error and not self.done:
            time.sleep(self._delay)

    def join(self, close=True, terminate=False, raise_exc=None, timeout=None):
        # close stream
        if close:
            self.close()
        if terminate:
            self.terminate()

        # join any sinks that need it
        for s in self.sinks:
            if hasattr(s, 'join'):
                s.join()

        # close thread
        if self._thread is not None:
            self._thread.join(timeout=timeout)
        if (self.controlling if raise_exc is None else raise_exc):
            self.raise_exception()
        if self._print_summary:  # print now or part of the stats will be lost if the block is used inside of Task
            print(self.stats_summary())

    def raise_exception(self):
        self._except.raise_any()

    def log_exception(self):
        for e in self._except.all():
            self.log.exception(e)

    @property
    def _exception(self):
        return self._except.last

    @property
    def all_exceptions(self):
        return self._except.all()

    def __export_state__(self):
        return {
            '_sw': self._sw,
            'started': self.started, 'ready': self.ready,
            'done': self.done, #'error': self.error,
            'terminated': self.terminated,
            # '_stream.terminated': self._stream.terminated,
            # '_stream.should_wait': self._stream.should_wait,
            # '_stream.running': self._stream.running,
            '_except': self._except,
        }

    def __import_state__(self, state):
        for k, v in state.items():
            try:
                x, ks = self, k.lstrip('!').split('.')
                for ki in ks[:-1]:
                    x = getattr(x, ki)
                setattr(x, ks[-1], v)
            except AttributeError:
                raise AttributeError('Could not set attribute {} = {}'.format(k, v))

    # State management

    def pause(self):
        if self._stream is not None:
            self._stream.pause()

    def resume(self):
        if self._stream is not None:
            self._stream.resume()

    def close(self, propagate=False):
        if self._stream is not None:
            self._stream.close()
        self.closed = True
        if propagate:
            self._send_sink_signal(reip.CLOSE)

    def terminate(self, propagate=False):
        if self._stream is not None:
            self._stream.terminate()
        if propagate:
            self._send_sink_signal(reip.TERMINATE)

    @property
    def error(self):
        return self._exception is not None

    # XXX: this is temporary. idk how to elegantly handle this
    @property
    def running(self):
        return self._stream.running if self._stream is not None else False

    # @property
    # def closed(self):
    #     return self._stream.should_wait if self._stream is not None else True

    @property
    def terminated(self):
        return self._terminated or (self._stream.terminated if self._stream is not None else False)

    @terminated.setter
    def terminated(self, value):
        self._terminated = value


    # debug

    def short_str(self):
        return '[B({})[{}/{}]({})]'.format(
            self.name, len(self.sources), len(self.sinks),
            self.block_state_name)

    def stats(self):
        total_time = self._sw.stats().sum if '' in self._sw else 0
        init_time = self._sw.stats("init").sum if 'init' in self._sw else 0
        finish_time = self._sw.stats("finish").sum if 'finish' in self._sw else 0

        return {
            'name': self.name,
            'total_time': total_time,
            'processed': self.processed,
            'speed': self.processed / (total_time - init_time - finish_time) if total_time else 0,
            'dropped': [getattr(s, "dropped", None) for s in self.sinks],
            'skipped': [getattr(s, "skipped", None) for s in self.sources],
            'n_in_sources': [len(s) if s is not None else None for s in self.sources],
            'n_in_sinks': [len(s) if s is not None else None for s in self.sinks],
            'sw': self._sw,
            'exception': self._exception,
            'all_exceptions': self._except._groups,
        }

    def summary(self):
        return text.block_text(
            text.green(str(self)),
            'Sources:',
            text.indent(text.b_(*(f'- {s}' for s in self.sources)) or None, w=4)[2:],
            '',
            'Sinks:',
            text.indent(text.b_(*(f'- {s}' for s in self.sinks)) or None, w=4)[2:],
            ch=text.blue('*'), n=40,
        )

    def status(self):
        '''
        e.g. `Block_123    +   2 buffers (2.09 x/s),   9 total (1.47 x/s avg),  sources=[1], sinks=[0]`
        '''
        n = self.processed
        total_time = self._sw.elapsed()
        init_time = self._sw.stats("init").sum if 'init' in self._sw else 0
        speed_avg = n / (total_time - init_time)

        n_new = n - self.old_processed
        speed_now = n_new / (time.time() - self.old_time)
        self.old_processed, self.old_time = n, time.time()

        n_src = [len(s) for s in self.sources]
        n_snk = [len(s) for s in self.sinks]

        return f'{self.name}\t + {n_new:3} buffers ({speed_now:,.2f} x/s), {n:5} total ({speed_avg:,.2f} x/s avg),  sources={n_src}, sinks={n_snk}'

    def stats_summary(self):
        stats = self.stats()
        # return text.block_text(
        return text.b_(
            # block title
            '\nStats for {summary_banner}',
            # any exception, if one was raised.
            text.red('({}) {}'.format(type(self._exception).__name__, self._exception))
            if self._exception else None,
            # basic stats
            'Processed {processed} buffers in {total_time:.2f} sec. ({speed:.2f} x/s)',
            'Dropped: {dropped}  Skipped: {skipped}  Left in Queue: in={n_in_sources} out={n_in_sinks}',
            # timing info
            # self._sw, ch=text.blue('*')).format(
            self._sw).format(
                summary_banner=text.red(self) if self.error else text.green(self),
                **stats)

    # def print_stats(self):
    #     print(self.stats_summary())


def prepare_output(outputs, input_meta=None, expected_length=None):
    '''Take the inputs from block.process and prepare to be passed to block.sinks.'''
    if not outputs:
        return (), {}

    bufs, meta = None, None
    if isinstance(outputs, tuple):
        if len(outputs) == 2:
            bufs, meta = outputs
    elif isinstance(outputs, (Meta, dict)):
        meta = outputs

    bufs = list(bufs) if bufs else []

    if expected_length:  # pad outputs with blank values
        bufs.extend((reip.BLANK,) * max(0, expected_length - len(bufs)))

    if input_meta:  # merge meta as a new layer
        if isinstance(meta, Meta):  # only the user did not wish to override it
            meta = Meta(meta, input_meta)
        else:
            meta = Meta(meta)
    return bufs, meta or Meta()



'''

Outstanding issues:
 - Every loop iteration must yield something otherwise it will hang
    - we may need asyncio for this kind of block :/
    - is there any way to

'''
import sys
import functools
from contextlib import contextmanager
import reip


class Streamer(reip.Block):
    _stream = _block = None
    def __init__(self, new, *a, name=None, **kw):
        self._new_block = new
        self._extra_args = a
        self.__class__ = type(new.__name__, (self.__class__,), {})
        super().__init__(name=name, extra_kw=True, **kw)

    def init(self):
        self._feeder = _feeder()
        next(self._feeder)  # clear initial value
        self._block = self._new_block(
            self._feeder, *self._extra_args, **self.extra_kw)

    def process(self, *xs, meta):
        # still assumes 1-to-1
        self._feeder.send(*xs, meta)
        return next(self._block)

    def finish(self):
        # run to completion and discard anything additional
        # QUESTION: should we catch exceptions here?
        # print('FINISHING', self, list(self._block))
        # for _ in self._block:
        #     pass
        self._block.close()


def _feeder(value=None):
    while True:
        value = yield value


def streamer(func=None):
    @functools.wraps(func)
    def outer(*a, **kw):
        return Streamer(func, *a, **kw)
    return outer


class _ProcessFunctionBlock(reip.Block):
    _BLOCK_INIT_ARGS = {}
    def __init__(self, *a, **kw):
        super().__init__(*a, **self._BLOCK_INIT_ARGS, **kw)

@reip.util.decorator
def process_func(func, **kw):
    return type(func.__name__, (_ProcessFunctionBlock,), {
        'process': func, '_BLOCK_INIT_ARGS': kw
    })


class _ContextFunc(reip.Block):
    _context = _process = None
    def __init__(self, func, **kw):
        self.func = contextmanager(func)
        super().__init__(extra_kw=True, **kw)

    def init(self):
        self._context = self.func(**self.extra_kw)
        self._process = self._context.__enter__()

    def process(self, *xs, meta):
        return self._process(*xs, meta)

    def finish(self):
        self._context.__exit__(*sys.exc_info())

def context_func(func=None):
    @functools.wraps(func)
    def outer(*a, **kw):
        return _ContextFunc(func, *a, **kw)
    return outer


# __name__ = '__main__'
if __name__ == '__main__':
    import reip.blocks as B

    @streamer
    def custom_block(stream=..., step=2):  # the ... is just to shut up pylint.
        try:
            # basically init() - do some initialization
            start = 10
            # basically process() - do some processing
            for i, ((data,), meta) in enumerate(stream, start):
                yield [data * step], {'index': i}
        finally:
            # basically finish() - do some cleanup
            print(f'\n\n!!!!!!!!!! all done. just cleaning up {stream}. !!!!!!!\n\n')


    (B.Interval(1)
     .to(B.Increment())
     .to(custom_block(step=4))  # no stream parameter was making pylint freak out
     .to(B.Debug('asdf')))
    reip.run(duration=6)

    @ContextFunc
    def block():
        try:
            ... # do_init()
            def process(x, meta):
                return [x], {}
            yield process
        finally:
            ... # do_finish()



import os
import tarfile
import time
import reip


class Tar(reip.Block):
    '''Archive files in a .tar file.'''
    def __init__(self, filename='{time}.tar', remove_files=False, gz=None, **kw):
        self.filename = filename
        self.remove_files = remove_files
        self.gz = filename.endswith('.gz') if gz is None else gz
        super().__init__(**kw)

    def process(self, *files, meta):
        # get filename
        meta["time"] = time.time()
        fname = self.filename.format(**meta)
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        # write to tar
        with tarfile.open(fname, "w:gz" if self.gz else 'w') as tar:
            for f in files:
                tar.add(f, arcname=os.path.basename(f))
        if self.remove_files:
            for f in files:
                os.remove(f)
        return [fname], {}

class TarGz(Tar):
    def __init__(self, filename='{time}.tgz', **kw):
        super().__init__(filename, gz=True, **kw)



'''Utils for managing wpa supplicant files.
'''
import os
import time
from shutil import copyfile
from sonycnode.utils import misc
from sonycnode.utils import sonyc_logger as loggerClass
from sonycnode.utils.settings import config, confuse


logger = loggerClass.sonyclogger(loggername="WpaSupplicant")

cfg = config['network'].get({
    'ap_path': confuse.Filename('/etc/wpa_supplicant/aps'),
    'wlan_name': 'wlan0',
})

REPO_AP_PATH = os.path.join(os.path.dirname(__file__), 'aps')
WPA_FNAME = "/etc/wpa_supplicant/wpa_supplicant.conf"
AP_PATH = cfg.ap_path
AP_PATTERN = os.path.join(cfg.ap_path, '{}.conf')

os.makedirs(cfg.ap_path, exist_ok=True)


def connect(ssid):
    return Wpa(ssid).connect()


class Wpa:
    def __init__(self, ssid=None, path=None):
        self.path = path or (Wpa.ssid_path(ssid) if ssid else WPA_FNAME)
        self.ssid = ssid or self.parsed.get('ssid')

    @misc.log_error_as_error(logger, msg="Error connecting to ap: ", default=False)
    def connect(self, backup=True):
        '''Set ap as current wpa_supplicant.'''
        wpa = Wpa()
        if wpa.ssid != self.ssid:
            if backup:
                wpa.backup()
            return self.copy(WPA_FNAME) and restart_iface(cfg.wlan_name)
        return True

    @misc.log_error_as_error(logger, msg="Error copying wpa_supplicant: ", default=False)
    def copy(self, dest):
        '''Copy wpa_supplicant to destination.'''
        if self.path != dest:
            copyfile(self.path, dest)
        return True

    def create(self, **kw):
        '''Generate wpa supplicant.'''
        generate_wpa_config(self.ssid, **kw)

    @misc.log_error_as_error(logger, msg="Error copying to wpa_supplicant: ", default=False)
    def backup(self, force=True):
        '''Make sure that the current wifi network is present in aps/.
        '''
        if force or self.ssid not in ssids_from_dir(cfg.ap_path):
            self.copy(self.ssid_path(self.ssid))
            return True
        return False

    @misc.cached_property
    def parsed(self):
        '''Parse the wpa_supplicant.conf file and return key value pairs.'''
        DROP_KEYS = ('network',)
        with open(self.path) as wpa_conf:
            return {
                x[0].strip(): x[1].strip('" ') for x in
                (l.split("=", 1) for l in wpa_conf.read().splitlines())
                if len(x) > 1 and x[0] not in DROP_KEYS
            }

    def __getattr__(self, k):
        try:
            return self.parsed[k]
        except KeyError as e:
            raise AttributeError(e)

    @staticmethod
    def ssid_path(ssid):
        return AP_PATTERN.format(ssid)




def ssids_from_dir(ap_path=cfg.ap_path, pat='*.conf'):
    '''Get file name -> file path mapping for files in a directory.
    e.g.: `{file: ap_path/file.ext for f in glob(ap_path)}`'''
    return {
        os.path.splitext(os.path.basename(f))[0]: f
        for f in misc.prodiglob(ap_path, pat)}

def sync_ap_directory(repo_path=REPO_AP_PATH, ap_path=cfg.ap_path, force=True, backup=True):
    '''Copy aps from network/aps to the trusted aps path.'''
    fnames_repo = ssids_from_dir(repo_path)
    fnames_aps = ssids_from_dir(ap_path)
    for fn in (fnames_repo if force else set(fnames_repo) - set(fnames_aps)):
        copyfile(fnames_repo[fn], os.path.join(ap_path, fn + '.conf'))
    if backup:
        Wpa().backup()





@misc.log_error_as_error(logger, msg="Error restarting interface: ")
def restart_iface(ifname=None, sleep=5):
    '''Restart the specified network interface. Returns True if restarted without error.'''
    logger.info("Restarting Interface")
    o, err = misc.execute('sudo ifdown {} --force && sleep {}'.format(ifname, sleep))
    #logger.info('ifdown: {} || {}'.format(o, err))
    time.sleep(sleep)
    o, err = misc.execute('sudo ifup {} && sleep {}'.format(ifname, sleep))
    #logger.info('ifup: {} || {}'.format(o, err))
    return True



'''

Create Wpa Supplicant File

'''


@misc.log_error_as_error(logger, msg="Error creating wpa_supplicant: ", default=False)
def generate_wpa_config(ssid, password=None, kind='basic', group='netdev', country='US', **kw):
    '''Create the wpa_supplicant configuration file
    in the `./aps/{ssid}.conf`

    Arguments:
        ssid (str): Access point to add credentials for.
        password (str, optional): The

    Returns:
        status (bool): True if no exception was raised when creating
    '''
    logger.debug("Creating config for: " + str(ssid))
    if kind == 'basic':
        if password:
            network = dict(ssid=ssid, psk=password)
        else:
            network = dict(ssid=ssid, key_mgmt='NONE')
    elif kind == 'wpa-eap':
        network = dict(
            ssid=ssid, proto='RSN', key_mgmt='WPA-EAP',
            pairwise='CCMP', phase2="auth=MSCHAPV2",
            auth_alg='OPEN', eap='PEAP',
            identity=kw.pop('identity'),
            password=password)
    else:
        raise ValueError('Unknown wpa config format "{}"'.format(kind))

    tmpl = misc.redent('''
        ctrl_interface=DIR=/var/run/wpa_supplicant GROUP={group}
        update_config=1
        country={country}
        network={
        {network}
        }'''.strip())
    with open(Wpa.ssid_path(ssid), 'w') as f:
        f.write(tmpl.format(
            group=group, country=country,
            network=misc.redent(_wpa_keys(**network, **kw), 2)))
    return True


def _wpa_keys(*a, sep='\n', **kw):
    return sep.join(
        list(map(str, a)) +
        ['{}={}'.format(k, '"{}"'.format(v) if not isinstance(v, (int, float)) else v)
         for k, v in kw.items()])

def askpass():
    import getpass
    return getpass.getpass()



from .base import BaseStore


class Store(BaseStore):
    '''A basic store that stores it's elements in a list.'''
    def __init__(self, size):
        self.items = [None] * size

    def __len__(self):
        return len(self.items)

    def put(self, data, meta=None, id=None):
        self.items[id] = data, meta

    def get(self, id):
        return self.items[id]

    def delete(self, ids):
        for i in ids:
            self.items[i] = None



import multiprocessing as mp
mp.set_start_method('fork')

import logging
log = logging.getLogger(__name__)
# log.propagate = True
log.setLevel(logging.INFO)

def package_file(*fs):
    return os.path.join(os.path.dirname(reip.__file__), *fs)

import time
from .constants import *
from . import exceptions
from . import util
from .util import status

from .interface import *
from . import stores
from .stores import Producer
from .stream import *
from .block import *
from .graph import *
from .task import *
Graph._initialize_default_graph()
from .helpers import *
from .blocks import *

def run(*a, **kw):
    default_graph().run(*a, **kw)

def graph_function(func):
    def inner(*a, **kw):
        with reip.Graph() as g:
            func(*a, **kw)
        g.run()
        return g
    return inner


def moment(delay=1e-6):
    '''Sleep for the minimum amount of time.'''
    time.sleep(delay)



import reip
import shlex
import subprocess



class Shell(reip.Block):
    '''Run a shell command on the input data. See :py:func:`reip.util.shell.run` for 
    information about command formatting.'''
    def __init__(self, cmd, astype=str, **kw):
        self.cmd = cmd
        self.astype = astype
        super().__init__(n_inputs=None, **kw)

    def process(self, *xs, meta):
        result = reip.util.shell.run(self.cmd, *xs, **meta)
        return [self.astype(result.out.strip())], {'stderr': result.err.strip()}


class ShellProcess(reip.Block):
    '''Spawn a shell process while this block is running. There is currently
    no support for passing data to and from the process, but their outputs 
    are available via self.stdout and self.stderr.
    '''
    _proc = _pid = None
    stdout = stderr = None
    n_err_lines = 50
    def __init__(self, cmd, **kw):
        super().__init__(**kw)
        self.cmd = cmd

    def init(self):
        self._proc = subprocess.Popen(
            shlex.split(self.cmd), stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.stdout = self._proc.stdout
        self.stderr = self._proc.stderr
        self._pid = self._proc.pid
        self.log.debug('Started process {}.'.format(self._pid))

    def finish(self):
        self.log.debug('Terminating process {}...'.format(self._pid))
        self._proc.terminate()
        self.log.debug('Waiting for process {} to finish...'.format(self._pid))
        self._proc.wait()
        # if self._proc.returncode:
        #     raise RuntimeError((
        #         'Shell process exited with return code {}. \n'
        #         '  command: {} \n\n'
        #         '  error (last {} lines): \n{}').format(
        #             self._proc.returncode, self.cmd, self.n_err_lines,
        #             '\n'.join(self.stderr.getvalue().decode('utf-8').splitlines()[-self.n_err_lines:]
        #         )))



'''

'''
import time
import threading
import multiprocessing as mp
from ctypes import c_bool
import remoteobj
from . import SharedPointer, Store, Customer, Queue


class _RemoteTraceback(Exception):
    def __init__(self, tb):
        self.tb = tb
    def __str__(self):
        return self.tb

class QueueCustomer(Customer):
    '''A customer that requests values from a Queue Store.'''
    cache = None
    def __init__(self, *a, serializer=None, **kw):
        super().__init__(*a, **kw)
        self.requested = mp.Value(c_bool, False, lock=False)
        self.data_queue = Queue(serializer)
        self.store.customers.append(self)  # circular reference - garbage collection issue?

    def empty(self):
        return self.store.quit.value or super().empty()

    def next(self):
        self.cache = None
        super().next()

    def _get(self):
        if self.cache is None:
            if not self.store.error.value:
                if self.store.quit.value:
                    raise RuntimeError('Store is not running.')
                self.requested.value = True
                v = self.data_queue.get()
            if self.store.error.value:
                exc = RuntimeError('Exception {} in {}'.format(v[0], self.store.__class__.__name__))
                exc.__cause__ = _RemoteTraceback(v[1])
                raise exc
            self.cache = v
        return self.cache




# class QueuePointer(SharedPointer):
#     cache = None
#     def __init__(self, size, counter=0, faster_queue=False):
#         super().__init__(size, counter)
#         self.requested = mp.Value(c_bool, False, lock=False)
#         self.data_queue = (
#             FasterSimpleQueue(ctx=mp.get_context()) if faster_queue else
#             mp.SimpleQueue()
#         )
#
#     def _get(self):
#         if self.cache is None:
#             self.requested.value = True
#             self.cache = self.data_queue.get()
#         return self.cache



class QueueStore(Store):
    '''A Store that will push values through a queue when a Customer requests.'''
    Pointer = SharedPointer
    Customer = QueueCustomer

    debug = False
    _thread = None
    def __init__(self, *a, **kw):
        self.customers = []
        self.quit = mp.Value(c_bool, False, lock=False)
        self.error = mp.Value(c_bool, False, lock=False)
        super().__init__(*a, **kw)

    def spawn(self):
        if self.debug:
            print("Spawning producer", self)
        self.quit.value = self.error.value = False
        self._thread = threading.Thread(target=self._run, daemon=True)
        self._thread.start()

    def join(self):
        if self._thread is not None:
            if self.debug:
                print("Joining producer", self)
            self.quit.value = True
            self._thread.join()
            self._thread = None
            if self.debug:
                print("Joined producer", self)

    def _run(self):
        if self.debug:
            print("Spawned producer", self)
        while not self.quit.value:
            for c in self.customers:
                if c.requested.value:
                    try:
                        c.requested.value = False
                        c.data_queue.put(self.items[c.cursor.pos])
                    except Exception as e:
                        self.error.value = True
                        import traceback
                        c.data_queue.put((type(e).__name__, traceback.format_exc()))
            time.sleep(1e-6)
        if self.debug:
            print("Exiting producer", self)



from ..interface import Source
from reip.util import text


class Customer(Source):
    '''A Source object to read from a :py:class:`reip.stores.producer.Producer`
    '''
    def __init__(self, source, index, store_id, **kw):
        self.source = source
        self.id = index
        self.store_id = store_id
        super().__init__(**kw)

    def __str__(self):
        return '<{}[{}] queued={} {} of \n{}>'.format(
            self.__class__.__name__, self.id, len(self),
            self.source.readers[self.id],
            text.indent(self.source))

    def __len__(self):
        return self.source.head.counter - self.cursor.counter

    def next(self):
        '''Increment this cursor.'''
        if len(self):
            self.cursor.counter += 1

    def _get(self):
        return self.store.get(self.cursor.pos)

    @property
    def cursor(self):
        '''This customer's reader cursor. It marks the index
        of the data within the Producer's Store.'''
        return self.source.readers[self.id]

    @property
    def store(self):
        '''Links to the corresponding store from the Producer.
        This will vary depending on the configuration across
        Tasks.
        '''
        return self.source.stores[self.store_id]



import time
import collections
import numpy as np
import reip
from reip.util.statistics import OnlineStats


_BLANK = ''

class Stopwatch:
    '''Stopwatch for timing and recording execution time of bits of code.
    
    .. code-block:: python

        sw = reip.util.Stopwatch()

        for _ in range(100):
            with sw('sleeping'):
                time.sleep(0.1)
    '''
    class lap:
        def __init__(self, sw, name=_BLANK):
            self.sw = sw
            self.name = name

        def __enter__(self):
            self.sw.tick(self.name)

        def __exit__(self, exc_type, exc_val, exc_tb):
            self.sw.tock(self.name)

    def __init__(self, title='', max_samples=500):
        self._max_samples = max_samples
        self._title = title
        self._dt = self._estimate_dt()[0]
        self.reset()

    # def __getstate__(self):
    #     return dict(self.__dict__, _samples={})

    def reset(self):
        # self._samples = {}
        self._ticks = {}
        self._stats = {}
        # self._sums = {}
        # self._counts = {}
        # self._stds = {}

    def __contains__(self, key):
        return key in self._stats

    def _estimate_dt(self, n=100):
        dts = []
        for i in range(n):
            t0 = time.time()
            t1 = time.time()
            dts.append(t1 - t0)
        dts = np.array(dts)
        dt = np.min(dts[dts > np.mean(dts) / 2])
        return dt, dts

    def tick(self, name=_BLANK):
        if name not in self._stats:
            self._stats[name] = OnlineStats(self._max_samples)
        self._ticks[name] = time.time()

    def tock(self, name=_BLANK, samples=True):
        # Correct for time.time() execution time (dt) and Lap class overhead (approx 1 us)
        t = (self._ticks[name], time.time() - self._dt - 1e-6)
        self._stats[name].append(max(0, t[1] - t[0]))
        # self._sums[name] += max(0, t[1] - t[0])
        # self._counts[name] += 1
        # if samples:
        #     s = self._samples[name]
        #     s.append(t)
        #     if len(s) > self._max_samples:
        #         self._samples[name] = s[len(s)//2:]

    def notock(self, name=_BLANK):
        self._ticks.pop(name, None)

    def ticktock(self, delta, name=_BLANK):
        self._stats[name].append(delta)
        # self._sums[name] += max(0, delta)
        # self._counts[name] += 1

    def iter(self, iterable, name=_BLANK, samples=True):
        self.tick(name)
        for item in iterable:
            self.tock(name, samples=samples)
            yield item
            self.tick(name)
        self.notock(name)

    def elapsed(self, name=_BLANK):
        if name in self._ticks.keys():
            return time.time() - self._ticks[name]
        else:
            print("Attention: No tick available right now for \"%s\"" % name)
            return 1e+6

    def last(self, name=_BLANK):
        ts = self._stats[name]
        return ts[-1][1] - ts[-1][0] if ts else None

    def avg(self, name=_BLANK):
        return self._stats[name].mean

    def total(self, name=_BLANK):
        return self._stats[name].sum

    def sleep(self, delay=1e-6, name="sleep"):
        self.tick(name)
        time.sleep(delay)
        self.tock(name, samples=False)

    def __call__(self, name=_BLANK):
        return self.lap(self, name)

    def __enter__(self):
        return self.lap(self).__enter__()

    def stats(self, name=_BLANK):
        if name not in self._stats:
            raise ValueError(f'Measurement unavailable for lap {name!r}')
        return self._stats[name]

    def __getitem__(self, key):
        return self.total(key)

    def __str__(self):
        total = self._stats[_BLANK].sum if _BLANK in self else 0

        if total and 'service' not in self._stats:
            added = sum(self._stats[k].sum for k in self._stats if k)
            service = self._stats['service'] = OnlineStats()
            service.append(total - added)

        name_width = max((len(k) for k in self._stats), default=0) + 2
        return ('Total {:.4f} sec in {}:\n'.format(total, self._title) if total else '') + (
            '\n'.join(
            ('    ({percent:5.2f}%) - {name:<{name_width}} '
             '(avg = {mean:.6f} ± {std:.6f}, n = {count:8,}) {total}').format(
                name=name, name_width=name_width,
                total=reip.util.human_time(stats.sum),
                percent=100. * stats.sum / total if total else 0,
                mean=stats.mean, std=stats.std, count=stats.count,
             )
            for name, stats in self._stats.items() if name
        ) or '-- no stats available --')

    def plot(self, cols=2, figsize=(16, 9)):
        import matplotlib.pyplot as plt
        print(self)
        plt.figure(figsize=figsize)
        for i, k in enumerate(self._stats, 1):
            plt.subplot(cols, int(np.ceil(len(self._stats[k].samples) / cols)), i)
            # plot distribution
            samples = np.array(self._stats[k].samples[k])
            plt.hist(samples[:, 1] - samples[:, 0], bins=100)
            # set title with stats
            total, avg, std, count = self.stats(k)
            plt.title(
                f'{k or "total"}: {total} sec '
                f'(avg = {avg:.6f} ± {std:.6f}, n = {count:.0f})')
        plt.suptitle(self._title)
        plt.tight_layout(rect=[0, 0.03, 1, 0.95])


if __name__ == '__main__':
    import matplotlib.pyplot as plt
    sw = Stopwatch("Test")

    dt, dts = sw._estimate_dt(100000)
    print("dt", dt)
    plt.figure("dt")
    plt.hist(dts, bins=200, range=[0, 1e-6])

    with sw():
        t0 = time.time()
        with sw("mini"):  # ~10 us overhead with ~1.5 us overestimate (corrected for)
            t1 = time.time()
            time.sleep(1e-6)
            t2 = time.time()
        t3 = time.time()

        tick = t1 - t0
        sleep = t2 - t1
        tock = t3 - t2
        total = t3 - t0
        print("tick", tick)
        print("sleep", sleep)
        print("tock", tock)
        print("total", total)
        print("service", tick + tock - 2*dt)
        print("job", sleep + dt)
        print("")

        with sw("init"):
            time.sleep(0.1)

        for i in range(10):
            with sw("process"):
                time.sleep(0.01)

        with sw("finish"):
            time.sleep(0.1)

    sw.plot()
    plt.show()



import os
import re
import contextlib
from functools import wraps
import numpy as np

def always(*a, **kw):
    return True

def never(*a, **kw):
    return False


def resize_list(lst, length, value=None):
    '''Resize a list to be a specific length.

    Example:
    >>> x = [1, 2]
    >>> assert resize_list(x, 5) == [1, 2, None, None, None]
    >>> assert resize_list(x, 5, 10) == [1, 2, 10, 10, 10]
    >>> assert resize_list(x, 4, lambda: x[0]) == [1, 2, 1, 1]  # some callable
    '''
    if length is None or length == -1 or len(lst) == length:
        return lst
    return (
        lst + [
            value() if callable(value) else value
            for i in range(max(0, length - len(lst)))
        ]
    )[:length]


def decorator(__func__=None, **kw):
    '''A convenience wrapper that allows you to define a decorator with optional
    initialization parameters.

    Example:
    >>> def my_decorator(func, **kw):
    ...     @functools.wraps(func)
    ...     def inner(*a, **kwi):
    ...         return func(*a, **dict(kw, **kw))
    ...     return inner
    >>> @my_decorator
    ... def asdf(x=10, y=15):
    ...     return x+y
    >>> @my_decorator(y=20)
    ... def asdf2(x=10, y=15):
    ...     return x+y
    >>> assert asdf() == 25 and asdf2() == 30
    '''
    def decorated(func):
        @wraps(func)
        def inner(*a, **kwi):
            return func(*a, **dict(kw, **kwi))
        return inner
    return decorated(__func__) if callable(__func__) else decorated


def partial(__partial_func__, *a, __name__=None, **kw):
    @wraps(__partial_func__)
    def inner(*ai, **kwi):
        return __partial_func__(*a, *ai, **dict(kw, **kwi))
    if __name__ is not None:
        inner.__name__ = __name__
    return inner

def create_partial(__partial_func__, *a, __name__=None, **kw):
    return partial(partial, __partial_func__, *a, __name__=__name__ or __partial_func__.__name__, **kw)


def resolve_call(func, *a, **kw):
    return func(*a, **kw) if callable(func) else func

def as_func(func):
    return func if callable(func) else (lambda: func)


def ensure_dir(fname):
    '''Make sure that the directory that this filename is in exists. Does nothing
    if this file is in the current working directory.'''
    parent = os.path.dirname(fname)
    if parent:  # ignore if parent is cwd
        os.makedirs(parent, exist_ok=True)
    return fname


def adjacent_file(file, *f):
    return os.path.abspath(os.path.join(os.path.dirname(file), *f))


def fname(file):
    '''Get file name. e.g. path/to/fileA.txt => fileA'''
    return os.path.splitext(os.path.basename(file))[0]


def write(fname, *lines, mode='r'):
    '''write to file.'''
    d = os.path.dirname(fname)
    if d and not os.path.exists(d):
        os.makedirs(d)
    with open(os.path.expanduser(fname), mode) as f:
        f.write('\n'.join(map(str, lines)))


def as_list(x):
    '''Convert or wrap value as a list.

    Examples:
    >>> assert as_list(5) == [5]
    >>> assert as_list('asdf') == ['asdf']
    >>> assert as_list((1, 2)) == [1, 2]
    >>> assert as_list([1, 2]) == [1, 2]
    '''
    return (
        x if isinstance(x, list)
        else list(x) if isinstance(x, tuple)
        else [x])

def is_iter(iterable):
    return not hasattr(iterable,'__len__') and hasattr(iterable,'__iter__')

def as_iterlike(x, like=(list, tuple, set, np.ndarray)):
    return x if isinstance(x, like) or is_iter(x) else [x]

def squeeze(x):
    '''If the input is a one-element list or tuple, take the first element.
    (removes an unnecessary container.)

    Examples:
    >>> assert squeeze([5]) == 5
    >>> assert squeeze((1,)) == 1
    >>> assert squeeze([1, 2]) == [1, 2]
    >>> assert squeeze('asdf') == 'asdf'
    >>> assert squeeze(('asdf',)) == 'asdf'
    >>> assert squeeze((1, 2)) == (1, 2)
    '''
    return x[0] if isinstance(x, (list, tuple)) and len(x) == 1 else x

def notnone(x):
    return x is not None

def filter_none(xs):
    return [x for x in xs if x is not None]


def separate(xs, *conditions, keep_fails=True):
    conditions = conditions or ((lambda x: x is not None),)
    outs = [[] for i in range(len(conditions)+1)]
    for x in xs:
        for i, c in enumerate(conditions):
            if c(x):
                outs[i].append(x)
                break
        else:
            outs[-1].append(x)
    return outs


def flatten(X, args=(), call=False, **kw):
    if isinstance(X, (list, tuple)):
        yield from (x for xs in X for x in flatten(xs, call=call, args=args, **kw))
    elif call and callable(X):
        yield from flatten(X(*args, **kw), call=call, args=args, **kw)
    else:
        yield X


def mergedict(*dicts, args=(), call=True, **kw):
    return _merge(flatten(dicts, args=args, call=call, **kw), call=call)

def _merge(flat, call=True):
    out = {}
    for d in flat:
        if not d:
            continue
        if not call and callable(d):
            return ([out] if out else []) + [d] + as_list(_merge(flat, call=call))
        out.update(d)
    return out


def matchmany(text, *patterns):
    return {
        k: v for m in (re.search(p, text) for p in patterns) if m
        for k, v in m.groupdict().items()}


@contextlib.contextmanager
def multicontext(*items):
    '''Use a variable set of context managers as one.'''
    with contextlib.ExitStack() as stack:
        yield [stack.enter_context(x) for x in items]


# @wraps(mergedict)
# def _mergedicts(dicts, *a, **kw):
#     out = {}
#     for d in flatten(dicts, *a, call=True, **kw):
#         out.update(d)
#     return out


# >     1d:  1d02h05m
# >    1hr:  1h03m04s
# > 10mins:  11m05s
# >  2mins:   2m05s
#         : 90.065s
#         :  2.065s

_TIME_BREAKS = [
    (60 * 60 * 24 * 7, '{w:.0f}w:{d:.0f}d:{h:.0f}h'.format),
    (60 * 60 * 24,     '{d:.0f}d:{h:.0f}h:{m:.0f}m'.format),
    (60 * 60,          '{h:.0f}h:{m:.0f}m:{s:.0f}s'.format),
    (60 * 10,          '{m:.0f}m:{s:.0f}s'.format),
    (60 * 1,           '{m:.0f}m:{s:.1f}s'.format),
    (0,                '{s:.3f}s'.format),
]

def _factor_time(t):
    wks, days = divmod(t, 60*60*24*7)
    days, hrs = divmod(t, 60*60*24)
    hrs, mins = divmod(hrs, 60*60)
    mins, secs = divmod(mins, 60)
    return {'w': wks, 'd': days, 'h': hrs, 'm': mins, 's': secs}

def human_time(secs):
    suffix = ' ago' if secs < 0 else ''
    secs = abs(secs)
    return '({}{})'.format(next((
        fmt(**_factor_time(secs))
        for t, fmt in _TIME_BREAKS if secs >= t
    )), suffix)


# class MpValueProp:
#     def __init__(self, name):
#         self.name = name
#
#     def __get__(self, instance, owner=None):
#         return getattr(instance, self.name).value
#
#     def __set__(self, instance, value):
#         getattr(instance, self.name).value = value



import numpy as np
import librosa
import reip


class ExampleAudio(reip.Block):
    y = sr = None
    def __init__(self, window=3, **kw):
        self.window = window
        super().__init__(**kw)

    def init(self):
        self.y, self.sr = librosa.load(librosa.util.example_audio_file())

    def process(self, meta=None):
        i = np.random.randint(len(self.y) - self.window * self.sr)
        y = self.y[i:i + self.window * self.sr]
        return [y], {
            'sr': self.sr,
            'offset': i / self.sr,
            'window': self.window
        }

    def finish(self):
        self.y = self.sr = None



import os
import sys
import time
import logging
import colorlog

MULTILINE_FORMAT = (
    '%(log_color)s[%(levelname)s]%(reset)s %(thin)s%(asctime)s%(reset)s: '
    '%(log_color)s%(message)s%(reset)s\n'
    '%(log_color)s %(block)s%(reset)s\n'
)
# [DEBUG] 2020-11-23 19:54:28,836: Starting...
#  [Block(Csv_3979680560): (1/1 in, 1 out; 0 processed) - --]

COMPACT_FORMAT = (
    '%(log_color)s[%(levelname)s]%(reset)s%(thin)s%(asctime)s%(reset)s '
    '%(log_color)s %(block)s%(reset)s: %(log_color)s%(message)s%(reset)s'
)
# [DEBUG]2020-11-23 19:54:28,836 [B(Csv_3979680560[1|1])--]: Starting...

DATE_FORMAT = "%m/%d/%y %H:%M:%S"

DEFAULT_LEVEL = os.getenv('REIP_LOG_LEVEL') or 'info'


def getLogger(block, level=DEFAULT_LEVEL, compact=True, propagate=False):
    is_block = block is not None and not isinstance(block, str)
    log = logging.getLogger(block.name if is_block else block or 'reip')

    if getattr(log, '_is_configured_by_reip_', False):
        return log
    log._is_configured_by_reip_ = True

    if is_block:
        log.addFilter(InjectData(dict(
            block=StrRep(block, 'short_str' if compact else None)
        )))
    log.propagate = propagate
    log.setLevel(aslevel(DEFAULT_LEVEL if level is None else level))
    formatter = colorlog.ColoredFormatter(COMPACT_FORMAT if compact else MULTILINE_FORMAT)
    formatter.converter = time.localtime
    return add_stdouterr(log, formatter=formatter, level=level)


def add_stdouterr(log, formatter=None, level='info', errlevel=logging.WARNING):
    # https://stackoverflow.com/questions/16061641/python-logging-split-between-stdout-and-stderr
    add_filehandle(log, sys.stdout, level, errlevel, formatter=formatter)
    add_filehandle(log, sys.stderr, errlevel, formatter=formatter)
    return log


def add_filehandle(log, file, minlevel=None, maxlevel=None, formatter=None):
    hand = levelrange(logging.StreamHandler(file), aslevel(minlevel), aslevel(maxlevel))
    if formatter is not None:
        hand.setFormatter(formatter)
    log.addHandler(hand)
    return log


def levelrange(log, minlevel=logging.DEBUG, maxlevel=None):
    if minlevel is not None:
        log.setLevel(aslevel(minlevel))
    if maxlevel is not None:
        log.addFilter(MaxLevel(aslevel(maxlevel)))
    return log

def aslevel(level):
    if str(level) == level:
        levelkey = level.upper() if level not in logging._nameToLevel else level
        return logging._nameToLevel.get(levelkey, level)
    return level

def minlevel(debug=False):
    return logging.DEBUG if debug else logging.INFO


class StrRep:
    '''Wrapping an object to provide an alternative string representation.'''
    def __init__(self, obj, method=None, *a, **kw):
        self._str = getattr(obj, method or '__str__')
        self.a, self.kw = a, kw

    def __str__(self):
        return self._str(*self.a, **self.kw)


class MaxLevel(logging.Filter):
    '''Filters (lets through) all messages with level < LEVEL'''
    def __init__(self, level):
        self.level = level

    def filter(self, record):
        return record.levelno < self.level


class InjectData(logging.Filter):
    def __init__(self, fields, *a, **kw):
        self._fields = fields
        super().__init__(*a, **kw)

    def filter(self, record):
        for k, v in self._fields.items():
            setattr(record, k, v)
        return True



# from pkg_resources import resource_filename
# import pooch
#
#
# _path = os.environ.get("LIBROSA_DATA_DIR", pooch.os_cache("librosa"))
# _DOGGY = pooch.create(
#     __data_path, base_url="https://librosa.org/data/audio/", registry=None
# )
#
# _DOGGY.load_registry(
#     resource_filename(__name__, str(Path("example_data") / "registry.txt"))
# )
#
# with open(
#     resource_filename(__name__, str(Path("example_data") / "index.json")), "r"
# ) as fdesc:
#     __TRACKMAP = json.load(fdesc)



'''

Install:

Conda:
$ conda install pygobject gtk3 gst-plugins-base


'''
import time
import reip
from reip.util import gstream

class GStreamer(reip.Block):
    _sample_timeout = 1e+6
    def __init__(self, stream, *outputs, lazy_sources=True):
        self._gs = stream
        self._outputs = outputs
        self._lazy_sources = lazy_sources
        super().__init__(n_inputs=0, n_outputs=len(outputs))

    def init(self):
        self._gs.start()
        while not self._gs.ready and not self.error and not self.done:
            time.sleep(self._delay)

    def process(self, meta):
        self._gs.check_messages()
        if self._gs.done:
            return reip.CLOSE
        if self._outputs:
            # will only pull samples if the sink has readers
            samples = [
                self._gs[o].try_pull_sample(self._sample_timeout)
                if not self._lazy_sources or self.sinks[i].readers else None
                for i, o in enumerate(self._outputs)]
            if all(samples):
                imgs, tss = zip(*(
                    gstream.unpack_sample(s)
                    if s is not None else (None, None)
                    for s in samples))
                return imgs, {'timestamps': tss}

    def finish(self):
        self._gs.end()

    @property
    def running(self):
        return self._gs.running

    def pause(self):
        self._gs.pause()
        super().pause()

    def resume(self):
        self._gs.resume()
        super().resume()



import time
import copy
import reip
from reip.util import text, iters


class Stream:
    '''Multiplexing multiple sources together. Abstracts away the internals of
    a block so that they can be used outside of a threaded context.


    with reip.Graph() as g:
        block = B.Block()

    with g.run_scope():
        # open a stream and iterate over it.
        with block.output_stream(duration=5) as stream:
            for (data,), meta in stream:
                print(data, meta)

    '''
    _retry = False
    _delay = 1e-6
    running = True
    terminated = False
    signal = None
    def __init__(self, sources, get_loop=None, auto_next=True, should_wait=True,
                 timeout=None, squeeze=False, name='', slice_key=None,
                 strategy=all, _sw=None, **kw):
        self.name = name or ''
        self.sources = sources
        self._loop_kw = kw
        self._get_loop = get_loop
        self.auto_next = auto_next
        self.should_wait = should_wait
        self.timeout = timeout
        self._squeeze = squeeze
        self._slice_key = slice_key
        self._sw = _sw or reip.util.Stopwatch(str(self))
        self._strategy = strategy


    def __str__(self):
        # srcs = ''.join(f'\n{s}' for s in self.sources)
        return '<{}({}) {} available={}>'.format(
            self.__class__.__name__, self.name,
            ' '.join((
                'will-wait' * self.should_wait,
                'terminated' * self.terminated))
            or ('running' if self.running else 'paused'),
            [len(s) for s in self.sources],
        )

    def get_loop(self, **kw):
        return (
            self._get_loop(**kw) if callable(self._get_loop) else
            self.default_loop(**dict(self._loop_kw, **kw)))

    @classmethod
    def default_loop(cls, duration=None, max_rate=None, delay=None):
        '''A master loop function that will contain'''
        return iters.throttled(
            iters.timed(iters.loop(), duration),
            max_rate, delay=delay or cls._delay)

    def __iter__(self):
        self._reset()
        for _ in self._sw.iter(self.get_loop(), 'sleep'):
            inputs = self.get()
            if self.terminated or (inputs is None and not self.should_wait):
                return
            if inputs is not None:
                if self._squeeze and len(self.sources) == 1:
                    inputs = inputs[0][0], inputs[1]
                if self._slice_key:
                    if self._slice_key == 'data':
                        inputs = inputs[0]
                    elif self._slice_key == 'meta':
                        inputs = inputs[1]
                yield inputs
                if self.auto_next:
                    self.next()
        # the only time this runs is when self.loop stops iterating.
        self.signal = reip.CLOSE

    def poll(self, block=True, timeout=None): # FIXME: return value? exception? ?????
        '''Returns whether or not the stream is ready to pull from.'''
        # for streams with no sources, there's nothing to wait for
        for t_asleep in self._sw.iter(iters.timed(iters.sleep_loop(), timeout or self.timeout), 'sleep'):
            if self.terminated:
                return False
            if not self.sources and not self.should_wait:
                return False
            if self.running and self._strategy(not s.empty() for s in self.sources):
                return True
            if not block or not self.should_wait:
                return False
        return False

    def get(self, block=True, timeout=None):
        # FIXME loop because if we see any signals, _get will be None and we'll have to
        #       try again. But we also
        for _ in iters.timed(iters.loop(), timeout or self.timeout):
            ready = self.poll(block=block, timeout=timeout)
            if not ready:  # either block=False or timeout
                return
            value = self._get()
            if value is not None:
                return value
            # otherwise it was a signal, retry

    def _get(self):
        with self._sw('source'):
            inputs = [s.get_nowait() for s in self.sources]
            # print(inputs)

            if inputs and all(reip.CLOSE.check(x[0]) for x in inputs if x is not None):
                self.signal = reip.CLOSE  # block will send to sinks
                self.next()
                self.close()
                return

            if inputs and any(reip.TERMINATE.check(x[0]) for x in inputs if x is not None):
                self.signal = reip.TERMINATE  # block will send to sinks
                self.next()
                self.terminate()
                return
            return prepare_input(inputs)

    def retry(self):
        self._retry = True
        return self

    def next(self):
        if not self._retry:
            for s in self.sources:
                s.next()
        self._retry = False
        return self

    def _reset(self):
        self.signal = None
        self.terminated = False
        self.resume()
        return self

    def open(self):
        self.should_wait = True
        self.terminated = False
        return self

    def close(self):
        self.should_wait = False
        return self

    def nowait(self, flag=True):
        self.should_wait = not flag
        return self

    def pause(self):
        self.running = False
        return self

    def resume(self):
        self.running = True
        return self

    def terminate(self):
        self.terminated = True
        return self

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()

    # pre-input

    def check_signals(self, outputs):
        if self.sources:
            if any(any(t.check(o) for t in reip.SOURCE_TOKENS) for o in outputs):
                # check signal values
                if len(outputs) > len(self.sources):
                    raise RuntimeError(
                        f'Too many signals for sources in {self}. '
                        f'Got {len(outputs)}, expected a maximum '
                        f'of {len(self.sources)}.')
                # process signals
                for s, out in zip(self.sources, outputs):
                    if out == reip.RETRY:
                        pass
                    else:
                        s.next()
                return True
        return False

    @classmethod
    def from_block_sources(cls, block, max_rate=None, auto_next=False, name=None, **kw):
        '''Generate a stream using a block's sources.

        NOTE: this is unsafe as a public interface, because if you're using
        the same source instances in multiple places, the calls to `.next()` will
        interfere and you'll end up skipping items. That is why `auto_next` is set
        to `False`. For now, this is just used internally when creating a block.
        '''
        return cls(
            block.sources, auto_next=auto_next,
            max_rate=max_rate or block.max_rate,
            name=name or block.name, **kw)

    @classmethod
    def from_block(cls, block, max_rate=None, duration=None, delay=None,
                   timeout=None, name=None, **kw):
        '''Generate a stream using sources generated from a block's sinks.'''
        return cls(
            [sink.gen_source(**kw) for sink in block.sinks],
            max_rate=max_rate or block.max_rate, delay=delay,
            duration=duration, timeout=timeout,
            name=name or block.name)  # FIXME: how to pass kw to Stream as well ...?

    @classmethod
    def from_sinks(cls, sinks, max_rate=None, duration=None, delay=None,
                   timeout=None, name=None, **kw):
        '''Generate a stream using sources generated from a block's sinks.'''
        return cls(
            [sink.gen_source(**kw) for sink in sinks],
            max_rate=max_rate, delay=delay,
            duration=duration, timeout=timeout,
            name=name)

    # Stream slicing

    '''Allows you to take a stream and select to only return a certain output
    or only the metadata.

    TODO: currently this will produce unexpected results if you were to iterate over
    multiple StreamSlices because they don't copy sources.
    A fix for this would be to generate a new source with its own Pointer objects.

    TODO: also, this could be optimized by not deserializing data if we are only requesting meta.

    Examples:
    >>> # creating 4 different stream slices
    >>> full = B.Constant(5).output_stream()
    >>> data = B.Constant(5).output_stream().data
    >>> data0 = B.Constant(5).output_stream().data[0]
    >>> meta = B.Constant(5).output_stream().meta

    >>> # ignore the fact that dicts are not hashable, this is for illustration.
    >>> with reip.default_graph().run_scope(duration=1):
    ...     assert set(full) == { ([5], {}) }  # everything, the default
    ...     assert set(data) == { [5] }        # all data buffers
    ...     assert set(full) == { 5 }          # only the first buffer
    ...     assert set(full) == { {} }         # only meta dicts
    '''

    @property
    def data(self):
        stream = copy.copy(self)
        stream._slice_key = 'data'
        return stream

    @property
    def meta(self):
        stream = copy.copy(self)
        stream._slice_key = 'meta'
        return stream

    def __getitem__(self, index):
        stream = copy.copy(self)
        stream.sources = reip.util.as_list(self.sources[index])
        stream._squeeze = stream._squeeze or not isinstance(index, slice)

        return stream


def prepare_input(inputs):
    '''Take the inputs from multiple sources and prepare to be passed to block.process.'''
    bufs = [buf for buf, meta in inputs] if inputs else ()
    metas = [meta for buf, meta in inputs] if inputs else []
    if len(metas) == 1:
        metas = metas[0]
    elif len(metas) == 0:
        metas = {}
    return bufs, metas

    # bufs, meta = zip(*inputs) if inputs else ((), ())
    # return (
    #     # XXX: ... is a sentinel for empty outputs - how should we handle them here??
    #     #      if there's a blank in the middle
    #     [b for bs in bufs for b in reip.util.as_list(bs) if b is not reip.BLANK],
    #     reip.util.Meta({}, *meta[::-1]))



import time
import remoteobj
import reip
from reip.util import iters, text


class Task(reip.Graph):
    _process = None
    _delay = 1e-3

    def __init__(self, *blocks, graph=None, **kw):
        super().__init__(*blocks, graph=graph, **kw)
        self.remote = remoteobj.Proxy(self, fulfill_final=True)
        self._except = remoteobj.Except()

    def __repr__(self):
        return self.remote.super.attrs_('__repr__')(_default=self.__local_repr)

    def __local_repr(self):
        self._pull_process_result()
        return super().__repr__()

    # main run loop

    def _run(self, duration=None, _controlling=True, _ready_flag=None):
        self.log.debug(text.blue('Starting'))
        try:
            with self._except(raises=False), self.remote.listen_(bg=False):
                try:
                    # initialize
                    super().spawn(wait=False, _controlling=_controlling, _ready_flag=_ready_flag)
                    while True:
                        if super().error or super().done:
                            return
                        if super().ready:
                            self.log.debug(text.green('Ready'))
                            break
                        self.remote.process_requests()
                        time.sleep(self._delay)

                    # main loop
                    for _ in iters.timed(iters.sleep_loop(self._delay), duration):
                        if super().done or super().error:
                            break
                        self.remote.process_requests()

                except KeyboardInterrupt as e:
                    self.log.info(text.yellow('Interrupting'))
                finally:
                    super().join(raise_exc=False)
        finally:
            _ = super().__export_state__()
            return _



    # process management
    def spawn(self, wait=True, _controlling=True, _ready_flag=None):
        if self._process is not None:  # only start once
            return
        self.controlling = _controlling

        self._reset_state()
        self._process = remoteobj.util.process(self._run, _ready_flag=_ready_flag, _controlling=_controlling)
        self._except = self._process.exc
        self._process.start()

        if wait:
            self.wait_until_ready()
        if self.controlling:
            self.raise_exception()

    def join(self, *a, timeout=10, raise_exc=True, **kw):
        if self._process is None:
            return

        self.log.debug(text.yellow('Joining'))
        self.remote.super.join(*a, raise_exc=False, _default=None, **kw)  # join children
        self._process.join(timeout=timeout, raises=False)
        self.__import_state__(self._process.result)

        self._process = None
        if raise_exc:
            self.raise_exception()
        self.log.debug(text.green('Done'))

    def _pull_process_result(self):
        # NOTE: this is to update the state when the remote process has finished
        if self._process is not None:
            r = self._process.result
            self.__import_state__(r)
            #self.log.info('result: {}'.format(r))

    def __export_state__(self):
        return self.remote.super.attrs_('__export_state__')(_default=None)

    # children state

    @property
    def ready(self):
        return remoteobj.get(self.remote.super.ready, default=False)

    @property
    def running(self):
        return remoteobj.get(self.remote.super.running, default=False)

    @property
    def terminated(self):
        return remoteobj.get(self.remote.super.terminated, default=self.__local_terminated)

    @property
    def done(self):
        return remoteobj.get(self.remote.super.done, default=self.__local_done)

    @property
    def error(self):
        return remoteobj.get(self.remote.super.error, default=self.__local_error)

    def __local_terminated(self):  # for when the remote process is not running
        self._pull_process_result()
        return super().terminated

    def __local_done(self):  # for when the remote process is not running
        self._pull_process_result()
        return super().done

    def __local_error(self):  # for when the remote process is not running
        self._pull_process_result()
        #self.log.debug('local error {} {}'.format(super().error, self))
        return super().error

    # block control

    def pause(self):
        return self.remote.super.pause()

    def resume(self):
        return self.remote.super.resume()

    def close(self):
        return self.remote.super.close(_default=None)

    def terminate(self):
        return self.remote.super.terminate(_default=None)

    # debug

    def stats(self):
        return self.remote.super.stats(_default=super().stats)

    def summary(self):
        return self.remote.super.summary(_default=super().summary)

    def status(self):
        return self.remote.super.status(_default=super().status)

    def stats_summary(self):
        return self.remote.super.stats_summary(_default=super().stats_summary)



import copy
import functools


def method_count(func):
    def inner(self, *a, **kw):
        try:
            count = self.method_counts
        except AttributeError:
            self.method_counts = count = {}
        count[name] = count.get(name, 0) + 1
        if isfunc:
            return func(self, *a, **kw)
    isfunc = callable(func)
    inner.__name__ = name = func.__name__ if isfunc else func
    return inner


# def count_calls(obj=None):
#     cls = obj.__class__
#     obj.method_call_counts = {}
#     obj.__class__ = type(cls.__name__, (_CallCount, cls,), {'__oldcls': cls})
#     return obj
#
# class _CallCount:
#     def revert(self):
#         self.__class__ = self.__oldcls
#
#     def __getattribute__(self, name):
#         if name != 'method_call_counts':
#             self.method_call_counts[name] = self.method_call_counts.get(name, 0) + 1
#         return super().__getattribute__(name)
#
#
# if __name__ == '__main__':
#     class A:
#         def a(self):
#             pass
#
#         def b(self):
#             pass
#
#         def __call__(self):
#             pass
#
#     a = count_calls(A())
#     a.a()
#     a.a()
#     a.b()
#     a()  # call doesn't work because access happens on the class
#
#
#     print(a.method_call_counts)



import multiprocessing as mp
import multiprocessing.queues as mpq
from multiprocessing import context

SERIALIZERS = {}

def get_serializer(name):
    '''Returns an object with a loads and dumps member.'''
    if name == 'pickle':
        return context.reduction.ForkingPickler
    if name == 'json':
        import json
        return json
    if name == 'ujson':
        import ujson
        return ujson
    if name == 'msgpack':
        import msgpack
        return msgpack
    if name in SERIALIZERS:
        return SERIALIZERS[name]
    if isinstance(name, str):
        raise ValueError('Serializer "{}" could not be found.'.format(name))
    return name

def register_serializer(cls, name=None):
    SERIALIZERS[name or cls.__name__] = cls


class _BytesQueue(mpq.SimpleQueue):
    def __init__(self, ctx=None):
        super().__init__(ctx=mp.get_context() if ctx is None else ctx)

    def get(self):
        with self._rlock:
            return self._reader.recv_bytes()

    def put(self, obj):
        if self._wlock is None:  # writes to win32 pipe are atomic
            self._writer.send_bytes(obj)
        else:
            with self._wlock:
                self._writer.send_bytes(obj)


class Queue(_BytesQueue):
    serializer = 'pickle'
    def __init__(self, serializer=None, **kw):
        self.serializer = get_serializer(serializer or self.serializer)
        super().__init__(**kw)

    def get(self):
        return self.serializer.loads(super().get())

    def put(self, obj):
        return super().put(self.serializer.dumps(obj))



import time
import os
import cv2
import reip
import tflit
import numpy as np

def _model_file(name):
    return os.path.join(os.path.dirname(reip.__file__), 'models', name)

class StyleTransfer(reip.Block):
    '''
    https://www.tensorflow.org/lite/models/style_transfer/overview#style_prediction
    '''
    MODEL_FILE = "magenta_arbitrary-image-stylization-v1-256_int8_transfer_1.tflite"
    STYLE_MODEL_FILE = "magenta_arbitrary-image-stylization-v1-256_int8_prediction_1.tflite"
    MODEL_IMAGE_DIM = 384
    STYLE_IMAGE_DIM = 256

    def __init__(self, style, **kw):
        super().__init__(**kw)
        self._init_style = style

    def init(self):
        self.model = tflit.Model(_model_file(self.MODEL_FILE))
        self.style_model = tflit.Model(_model_file(self.STYLE_MODEL_FILE))
        self.size = self.model.input_shape[1:-1]
        self.set_style(self._init_style)

    def process(self, frame, meta):
        X = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        X = preprocess_image(X, self.MODEL_IMAGE_DIM)

        styled = self.model.predict([X[None, ...], self._style_emb])[0]
        # undo
        # styled = cv2.resize(styled, frame.shape[:2][::-1])
        styled = np.uint8(styled * 127.5 + 127.5)
        styled = cv2.cvtColor(styled, cv2.COLOR_RGB2BGR)
        return [restore_aspect(styled, frame)], {}


    def set_style(self, img, img2=None, blend=0.5):
        self._style_emb = (
            self.blend_style(img, img2, blend)[0]
            if img2 is not None else
            self.generate_style(img)[0])
        return self._style_emb

    def blend_style(self, imgA, imgB, blend=0.5):
        styleA, imgA = self.generate_style(imgA)
        styleB, imgB = self.generate_style(imgB)
        return (blend * styleA + (1 - blend) * styleB), imgA, imgB

    def generate_style(self, img):
        if isinstance(img, str):
            img = cv2.imread(img, cv2.IMREAD_COLOR)
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        X = preprocess_image(img, self.STYLE_IMAGE_DIM)

        style_emb = self.style_model.predict(X[None, ...])
        return style_emb, img


def preprocess_image(img, target_dim=256):
    # Resize the image so that the shorter dimension becomes 256px.
    shape = img.shape[:2]
    shape = (np.asarray(shape) * target_dim / min(shape)).astype(int)
    img = cv2.resize(img, tuple(shape))

    # Central crop the image.
    img = resize_with_crop_or_pad(img, target_dim, target_dim, 3)
    img = (np.float32(img) - 127.5) / 127.5
    return img

def resize_with_crop_or_pad(arr, *size):
    out = np.zeros(size)
    diffs = [(s1 - s2) / 2 for s1, s2 in zip(arr.shape, size)]
    oslc = tuple(
        slice(int(max(0, -d)), int(min(s, s + d)))
        for d, s in zip(diffs, size))
    aslc = tuple(
        slice(int(max(0, d)), int(min(s, s - d)))
        for d, s in zip(diffs, arr.shape))
    out[oslc] = arr[aslc]
    return out

def restore_aspect(modified, original, stretch=False):
    mh, mw = modified.shape[:2]
    oh, ow = original.shape[:2]
    sh = mw * oh / ow
    sw = mh * ow / oh
    return cv2.resize(modified, (
        sh if (sh > mh) == bool(stretch) else mh,
        sw if (sw > mw) == bool(stretch) else mw,
    ))



import os
import glob
import shlex
import hashlib
import psutil
import subprocess
import atexit


class BackgroundServer:
    '''Sometimes a script needs some server running in the background.

    For example, Plasma needs to start a Plasma Store server process,
    but you only need one. Instead of requiring a user to manually do this,
    perhaps we can handle this automatically. Ideally, if a user has started
    their own server process, this should be able to identify and use the
    existing process.

    This will:
     - on spawn:
         - check if a server is already running.
         - if not, create the server
         - touch a file to let other processes know you're still using the server
     - on join:
         - remove the touched file
         - check if any other processes are still using the server.
         - if not, send SIGTERM to the server

    Example:
    >>> server = ServerProcess(
    ...     'plasma_store -s $TMPDIR/plasma -m 1000000000',
    ...     'plasma-store')

    >>> with server:
    ...     pa.put(data)


    '''
    SIGNAL = 15
    _server_pid = None
    managed = False
    def __init__(self, command, pattern=None):
        self.command = command
        self.pattern = pattern

        # tracking which processes are connected using a file in a directory.
        cmd_id = hashlib.md5(command.encode('utf-8')).hexdigest()[:9]
        self._pid_dir = os.path.join(
            os.getenv('TMPDIR', '/tmp'),
            'background_server_proc_{}'.format(cmd_id))

        # this process+instance's file. this means that you can have multiple
        # instances in a certain process and it won't conflict.
        self._pid_file = os.path.join(
            self._pid_dir, '{}__{}.pid'.format(str(id(self)), os.getpid()))

        # make sure the server pid collection dir exists.
        os.makedirs(self._pid_dir, exist_ok=True)

    @property
    def is_alive(self):
        '''Check if the server process is alive.'''
        return (
            self._server_pid in (p.pid for p in psutil.process_iter())
            if self._server_pid else self.search_process() is not None)

    @property
    def clients(self):
        '''List other clients.'''
        return os.listdir(self._pid_dir)

    def spawn(self):
        ctl_file = os.path.join(self._pid_dir, 'managed')
        self.managed = os.path.isfile(ctl_file)
        # check if a process already exists - if not, start a process
        p = self.search_process()
        if p is None:
            if self.managed:  # controlled process dead - swap to managed
                os.remove(ctl_file)
                self.managed = False
            p = subprocess.Popen(self.command, shell=True)
        elif not self.managed and not glob.glob(os.path.join(self._pid_dir, '*.pid')):
            touch(ctl_file)
            self.managed = True
        # store the process id
        self._server_pid = p.pid
        if self.managed:
            return
        # create the file
        touch(self._pid_file)
        atexit.register(self.join, atexit_=True)

    def join(self, atexit_=False):
        if not atexit_:
            atexit.unregister(self.join)
        if self.managed:
            return
        # remove our pid file and see who else is connected
        os.remove(self._pid_file)
        pids = {f.split('__')[-1].split('.')[0] for f in self.clients}

        # Check if all processes are dead. if yes, kill server
        alive_pids = {p.pid for p in psutil.process_iter()} & pids
        if not alive_pids:
            os.kill(self._server_pid, self.SIGNAL)

    def search_process(self):
        '''Match a process based on either the process name or the command string.'''
        return next((
            p for p in psutil.process_iter()
            if (
                # check for a substring in the process name
                self.pattern in p.name() if self.pattern else
                # if no substring was provided, then compare the commands
                shlex.split(self.command) == p.cmdline()
            )), None)

    def __enter__(self):
        self.spawn()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.join()


def touch(fname):
    with open(fname, 'a') as f:
        os.utime(f)



import os
import cv2
import reip
import tflit
import numpy as np


class Deeplabv3(reip.Block):
    '''
    Object detection model from tensorflow. 80 classes.

    https://github.com/EdjeElectronics/TensorFlow-Lite-Object-Detection-on-Android-and-Raspberry-Pi/blob/master/TFLite_detection_webcam.py

    Outputs:
      - boxes - Bounding box coordinates of detected objects
      - classes - Class index of detected objects
      - scores - Confidence of detected objects
      - num - Total number of detected objects (inaccurate and not needed)

    '''
    MODEL_FILE = "coco_ssd_mobilenet_v1_1.0_quant_2018_06_29.tflite"
    LABEL_FILE = "coco_ssd_mobilenet_v1_1.0_quant_2018_06_29_labelmap.txt"
    def __init__(self, draw=None, **kw):
        self._should_draw = draw
        super().__init__(n_outputs=2, **kw)

    def init(self):
        self.model = Deeplabv3Model()
        self.should_draw = bool(self.sinks[1].readers) if self._should_draw is None else self._should_draw
        self.log.debug(f'will draw: {self.should_draw}')

    def process(self, frame, meta):
        X = self.model.predict(frame[None])
        return [X, self.model.draw(frame, X[0]) if self.should_draw else None], meta


class Deeplabv3Model(tflit.Model):
    input_size = (257, 257)
    MODEL_FILE = "lite-model_deeplabv3_1_metadata_2.tflite"
    def __init__(self) -> None:
        self.size = self.input_shape[1:-1]
        super().__init__(reip.package_file('models', self.MODEL_FILE))

    def predict_batch(self, im, *a, **kw):
        im = cv2.resize(im[0][0], self.input_size)
        im = (im / 255).astype('float32')[None]
        return super().predict_batch([im], *a, **kw)

    def draw(self, im, segment, mix=0.5, xpad=5, rect_size=20, ypad=30, label_spacing=40):
        seg_cls = np.argmax(segment, axis=-1)
        ids, counts = np.unique(seg_cls, return_counts=True)
        seg_im = np.array(colors[seg_cls], dtype='float32')
        total = seg_cls.size

        if im.shape != seg_im.shape:
            seg_im = cv2.resize(seg_im, im.shape[:2][::-1])
        im = im * (1-mix) + seg_im * mix

        for i, (c, k) in enumerate(sorted(zip(counts[ids != 0], ids[ids != 0]), reverse=True)):
            if c > total * 0.01:
                color = tuple(map(int, colors[k]))
                cv2.rectangle(
                    im, 
                    (xpad, ypad + i * label_spacing - rect_size//2),
                    (xpad + rect_size, ypad + i * label_spacing + rect_size//2),
                    color, cv2.FILLED)
                label = f'{classes[k]} ({c / total:.2%})'
                txy = (xpad * 2 + rect_size, ypad + i * label_spacing)
                cv2.putText(im, label, txy, cv2.FONT_HERSHEY_SIMPLEX, 0.7, (0, 0, 0), 4)
                cv2.putText(im, label, txy, cv2.FONT_HERSHEY_SIMPLEX, 0.7, color, 2)
                
        return im / 255

Segment = Deeplabv3


classes = [
    '__background__', 'aeroplane', 'bicycle', 'bird', 'boat', 'bottle', 'bus',
    'car', 'cat', 'chair', 'cow', 'diningtable', 'dog', 'horse', 'motorbike',
    'person', 'pottedplant', 'sheep', 'sofa', 'train', 'tvmonitor']

colors = np.array([
    [0, 0, 0],
    [255,102,51],
    [255,179,153],
    [255,51,255],
    [255,255,153],
    [0,179,230],
    [230,179,51],
    [51,102,230],
    [153,153,102],
    [153,255,153],
    [179,77,77],
    [128,179,0],
    [128,153,0],
    [230,179,179],
    [102,128,179],
    [102,153,26],
    [255,153,230],
    [204,255,26],
    [255,26,102],
    [230,51,26],
    [51,255,204],
    [102,153,77],
    [179,102,204],
    [77,128,0],
    [179,51,0],
    [204,128,204],
    [102,102,77],
    [153,26,255],
    [230,102,255],
    [77,179,255],
    [26,179,153],
    [230,102,179],
    [51,153,26],
    [204,153,153],
    [179,179,26],
    [0,230,128],
    [77,128,102],
    [128,153,128],
    [230,255,128],
    [26,255,51],
    [153,153,51],
    [255,51,128],
    [204,204,0],
    [102,230,77],
    [77,128,204],
    [153,0,179],
    [230,77,102],
    [77,179,128],
    [255,77,77],
    [153,230,230],
    [102,102,255],
], dtype='float32')


import collections
import reip
# import time


class CycledWriter:
    '''Get rid of the boilerplate needed to

    '''
    _fname = None
    _writer = None
    def __init__(self, filename, file_length=60):
        self.filename = str(filename)
        self.file_length = file_length
        assert self.file_length is None or self.file_length > 0
        self.frame_index = 0
        self._closed_files = collections.deque()

    def get(self, meta):
        if self._writer is None or self._should_cycle(meta):
            self._writer = self._get_new_writer(meta)
        return self._writer

    def increment(self, n=1):
        self.frame_index += n

    def output_closed_file(self, meta=None):
        if self._closed_files:
            return [self._closed_files.popleft()], meta or {}

    def close(self):
        if self._writer is not None:
            self._close_writer()
        if self._fname:
            self._closed_files.append(self._fname)
        return self._fname

    def _should_cycle(self, meta):
        return self.file_length and self.frame_index >= self.file_length

    def _get_new_writer(self, meta):
        self.close()
        # get filename
        # meta["time"] = time.time()
        self._fname = fname = self.filename.format(**meta)
        reip.util.ensure_dir(fname)
        # make writer
        self.frame_index = 0
        return self._new_writer(fname, meta)

    def _new_writer(self, fname, meta):
        raise NotImplementedError

    def _close_writer(self):
        raise NotImplementedError



import time
import numpy as np
import reip


class Array(reip.Block):
    exec_mode = 'process'
    def __init__(self, shape, **kw):
        self.shape = shape
        self.array = None
        super().__init__(n_inputs=0, **kw)

    def init(self):
        self.array = np.ones(self.shape, dtype=np.uint8)

    def process(self, meta):
        return [self.array], {'shape': self.shape, 'i': self.processed}

    def finish(self):
        self.array = None
SomeArray = Array


class Op(reip.Block):
    def __init__(self, offset=0, **kw):
        self.offset = offset
        super().__init__(n_inputs=None, **kw)

    def process(self, *data, meta=None):
        return data, {'offset': self.offset}
SomeTransform = Op

class TimeBomb(reip.Block):
    clock = None
    def __init__(self, t_minus=5, min_rate=0.001, **kw):
        self.t_minus = t_minus
        super().__init__(n_inputs=None, min_rate=min_rate, **kw)

    def init(self):
        self.clock = time.time()

    def process(self, *xs, meta):
        dt = self.t_minus - (time.time() - self.clock)
        if dt <= 0:
            raise TimeoutError('boom!')
        return [dt], {}


class TextFile(reip.Block):
    def __init__(self, content, fname=None, **kw):
        self.content = (
            content if callable(content) else str(content or '').format)
        self.fname = str(fname)
        super().__init__(n_inputs=None, **kw)

    def process(self, *xs, meta):
        fname = self.fname.format(*xs, **meta)
        with open(fname, 'w') as f:
            f.write(self.content(*xs, **meta))
        return [fname], {}



'''


Usage:
>>> class MyObj:
...     def __init__(self):
...         self.remote = RemoteProxy(self)
...         self.stopped = False
...         self.data = []
...
...     def stop(self):
...         self.stopped = True
...
...     def get_data(self):
...         while not self.stopped:
...             self.data.append(time.time())
...             time.sleep(1)

>>> obj = MyObj()
>>> p = mp.Process(target=obj.get_data)
>>> p.start()
>>> time.sleep(3)
>>> obj.remote.stop()
>>> p.join()

'''
import time
import random
import threading
import traceback
import warnings
import multiprocessing as mp
import reip

__all__ = ['RemoteProxy']


FAIL_UNPICKLEABLE = False
SELF = reip.make_token('self')


def retrieve(view, **kw):
    return view.retrieve(**kw)


class RemoteView:
    '''Represents a set of operations that can be captured, pickled, and
    applied to a remote object.

    This supports things like:
        - `view.some_attribute`
            NOTE: this doesn't work with private or magic attributes
        - `view['some key']`
        - `view(1, 2, x=10)`
        - `view.some_method(1, 2)`
        - `view.super.some_method(1, 2)`
                (translates to `super(type(obj), obj).some_method(1, 2)`)

    '''
    _keys, _frozen = (), False
    def __init__(self, *keys, frozen=False):
        self._keys = keys
        self._frozen = frozen

    def __str__(self):
        '''Return a string representation of the Op.'''
        x = '?'
        for kind, k in self._keys:
            if kind == '.':
                if k == 'super':
                    x = f'super({x})'
                else:
                    x = f'{x}.{k}'
            elif kind == '[]':
                x = f'{x}[{x!r}]'
            elif kind == '()':
                args = ', '.join(
                    [f'{a:!r}' for a in k[0]] +
                    [f'{ka}={a!r}' for ka, a in k[1].items()])
                x = f'{x}({args})'
        return f'({x})'

    def resolve_view(self, obj):
        '''Given an object, apply the view - get nested attributes, keys, call, etc.'''
        orig = obj
        for kind, k in self._keys:
            if kind == '.':
                if k == 'super':
                    obj = super(type(obj), obj)
                else:
                    obj = getattr(obj, k)
            elif kind == '[]':
                obj = obj[k]
            elif kind == '()':
                obj = obj(*k[0], **k[1])
        return obj

    def _extend(self, *keys, **kw):
        '''Return a copy of self with additional keys.'''
        return RemoteView(*self._keys, *keys, **kw)

    def __getattr__(self, name):
        '''Append a x.y op.'''
        if name.startswith('_') or self._frozen:
            raise AttributeError(name)
        return self._extend(('.', name))

    def __getitem__(self, index):
        '''Append a x[1] op.'''
        if self._frozen:
            raise KeyError(index)
        return self._extend(('[]', index))

    def __call__(self, *a, **kw):
        '''Append a x(1, 2) op.'''
        if self._frozen:
            raise TypeError(f'{self} is frozen and is not callable.')
        return self._extend(('()', (a, kw)))


class RemoteProxy(RemoteView):
    '''Capture and apply operations to a remote object.

    Usage:


    '''
    _thread = None
    _delay = 1e-5
    def __init__(self, instance, *a, default=..., **kw):
        super().__init__(*a, **kw)
        self._obj = instance
        self._results = {}
        # self._local, self._remote = mp.Pipe()
        self._local, self._remote = mp.SimpleQueue(), mp.SimpleQueue()
        self._local_name = mp.current_process().name
        self._listening = mp.Value('i', 0, lock=False)
        self._default = default

    def __str__(self):
        return f'<Remote {self._obj} : {super().__str__()}>'

    def __getstate__(self):
        '''Don't pickle queues, locks, and shared values.'''
        return dict(self.__dict__, _thread=None, _listening=False,
                    _local=None, _remote=None)

    @property
    def is_local(self):
        '''Is the current process the main process or a child one?'''
        return mp.current_process().name == self._local_name

    # internal view mechanics. These override RemoteView methods.

    def _extend(self, *keys, **kw):
        # Create a new remote proxy object. Don't call RemoteProxy.__init__
        obj = self.__class__.__new__(self.__class__)
        obj.__dict__.update(self.__dict__)
        RemoteView.__init__(obj, *self._keys, *keys, **kw)
        return obj

    def _as_view(self):
        '''Convert to a view so we don't pickle the object instance, just the keys.'''
        return RemoteView(*self._keys, frozen=True)

    def __call__(self, *a, default=..., **kw):
        '''Automatically retrieve when calling a function.'''
        # You can no longer chain functions. They automatically retrieve.
        return super().__call__(*a, **kw).retrieve(default=default)

    # remote calling interface

    def poll_until_clear(self):
        '''Poll until the command queue is empty.'''
        while not self._remote.empty():
            self.poll()

    def poll(self):
        '''Check for and execute the next command in the queue, if available.'''
        if not self._remote.empty():
            # get the called function
            result_id, view = self._remote.get()
            try:  # call the function and send the return value
                result = view.resolve_view(self._obj)
            except BaseException as e:  # send any exceptions
                print(
                    f'An exception was raised in {self}.',
                    traceback.format_exc(), flush=True)
                self._local.put((result_id, None, e))

            try:
                if result is self._obj:  # solution for chaining
                    result = SELF
                self._local.put((result_id, result, None))
            except RuntimeError as e:
                if FAIL_UNPICKLEABLE:
                    warnings.warn(f'Return value of {view} is unpickleable.')
                    raise
                warnings.warn(
                    f"You tried to send an unpickleable object returned by {view} "
                    "via a pipe. We'll assume this was an oversight and "
                    "will return `None`. The remote caller interface "
                    "is meant more for remote operations rather than "
                    "passing objects. Check the return value for any "
                    "unpickleable objects. "
                    f"The return value: {result}")

    # parent calling interface

    def retrieve(self, default=...):
        if not self.is_local:  # if you're in the main thread, just run the function.
            return self.resolve_view(self._obj)
        if self.listening:  # if the remote process is listening, run
            # send command and wait for response
            result_id = random.randint(0, 10000000)
            view = self._as_view()

            self._remote.put((result_id, view))
            result = self.get_result(result_id)

            # if no result was returned, the queue was probably closed before
            # being able to execute the request.
            if result is not None:  # valid results are tuples
                # raise any exceptions returned
                x, exc = result
                if result == SELF:
                    result = self
                if exc is not None:
                    raise Exception(f'Remote exception in {self._obj}') from exc
                return x

        # if a default value is provided, then return that, otherwise return a default.
        default = self._default if default is ... else default
        if default is ...:
            raise RuntimeError(f'Remote instance is not running for {self}')
        return default

    def get_result(self, expected_id, wait=True):
        while self.listening:
            if not self._local.empty():
                out_id, x, exc = self._local.get()
                assert expected_id == out_id  # safety check
                return x, exc
            if not wait:
                break
            time.sleep(self._delay)


    # def get_result(self, result_id, wait=True):
    #     '''Get the result from the queue. If there are overlapping calls, and
    #     if you pull the result meant for a different thread, save it to
    #     the results dictionary and keep pulling results until you get
    #     your result id.
    #     '''
    #     # makes overlapping calls concurrency-safe
    #     while result_id not in self._results and self.listening:
    #         if not self._local.empty():
    #             out_id, x, exc = self._local.get()
    #             if out_id is result_id:
    #                 return x, exc
    #             self._results[out_id] = x, exc
    #         if not wait:
    #             break
    #         time.sleep(self._delay)
    #
    #     return self._results.pop(result_id, None)

    ########################################################################
    # If we don't want to poll for commands in the background and we
    # want all commands to happen in the main thread, then we can remove
    # the methods below.
    ########################################################################

    # running state - to avoid dead locks, let the other process know if you will respond

    def wait_until_listening(self, timeout=None):
        t0 = time.time()
        while not self.listening:
            if timeout and time.time() - t0 > timeout:
                raise TimeoutError('Waiting for remote {} instance to listen timed out.'.format(self))
            time.sleep(self._delay)

    @property
    def listening(self):
        '''Is the remote instance listening?'''
        return bool(self._listening.value)

    @listening.setter
    def listening(self, value):
        self._listening.value = int(value)

    def __enter__(self):
        self.listening = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.listening = False


    # remote background listening interface

    # def listen(self, block=False):
    #     '''Start listening. By default, this will launch in a background thread.'''
    #     # print('Starting listening...', self, flush=True)
    #     return self._run() if block else self._spawn()
    #
    # def _spawn(self):
    #     if self._thread is None:
    #         self._thread = threading.Thread(target=self._run)
    #         self._thread.daemon = True
    #         self._thread.start()
    #     return self
    #
    # def _join(self):
    #     self.listening = False
    #     if self._thread is not None:
    #         self._thread.join()
    #         self._thread = None
    #     return self
    #
    # def _run(self):
    #     self.listening = True
    #     while self.listening:
    #         self.poll()
    #         time.sleep(self._delay)
    #     self.listening = False
    #     time.sleep(0.1)
    #     self.poll() # one last check in case of race condition? idk may delete.
    #
    # def __enter__(self):
    #     return self.listen()
    #
    # def __exit__(self, exc_type, exc_val, exc_tb):
    #     self._join()


# def _find_type_matching(module_name):
#     objs = {}
#     for



'''

Tests

'''


class _A:
    x = 10
    term = False
    def __init__(self):
        self.remote = RemoteProxy(self)

    def __str__(self):
        return f'<A x={self.x} terminated={self.term}>'

    def asdf(self):
        self.x *= 2
        return self

    @property
    def zxcv(self):
        return self.x / 5.

    def terminate(self):
        self.term = True
        return self

    def start(self):
        self.term = False
        return self

class _B(_A):
    @property
    def zxcv(self):
        return self.x / 10.


if __name__ == '__main__':
    def test(a):
        print('str format')
        print(a.remote)
        print(a.remote.x)
        print(a.remote.asdf().asdf().zxcv)
        print()

        # test basic attribute
        print('attribute')
        print(a.remote.x._, a.x, 10, a.remote.x)
        print()
        # test attribute, call, return self
        print('remote update')
        print(a.remote.asdf())
        print(a.remote.asdf())
        print(a.remote.x.retrieve(), a.x, 20, 20, a.remote.x)
        print()

        print('local update')
        print(a.remote.asdf())
        print(a.asdf())
        print(a.remote.x.retrieve(), a.x, 40, 40, a.remote.x)
        print()
        # test property accessing attribute works
        print('property')
        print(a.remote.zxcv.retrieve(), a.zxcv, 4., 4., a.remote.zxcv)
        print()

        print('property')
        print(a.remote.super.zxcv.retrieve(), super(type(a), a).zxcv, 8., 8., a.remote.super.zxcv)
        print()

        print('local update')
        a.terminate()
        print(a.remote.term.retrieve(), a.term, False, True)
        a.start()
        print(a.remote.term.retrieve(), a.term, False, False)
        print()

        print('remote update')
        a.remote.terminate()
        print(a.remote.term.retrieve(), a.term, True, False)
        a.remote.start()
        print(a.remote.term.retrieve(), a.term, False, False)
        print()
        print(flush=True)


    def _run(obj):  # some remote job
        # obj.remote.listen(block=True)
        while True:
            obj.remote.poll()
            time.sleep(0.0001)

    obj = _B()

    p = mp.Process(target=_run, args=(obj,), daemon=True)
    p.start()
    test(obj)
    # time.sleep(3)
    p.terminate()
    p.join()



import os
import inspect
from . import text


def check_block(block, match='', *a):
    '''Only print out only blocks that match the query text. For debugging specific subclasses.'''
    if (not match or match.lower() in block.__class__.__name__.lower()
            or match.lower() in getattr(block, 'name', '').lower()):
        text.printasone(text.block_text(str(block), text.l_(*a)), flush=True)


def block_stack(message=None, fn=None, offset=0, limit=None):
    '''Format the current stack trace.'''
    stack = inspect.stack()
    f = stack[offset+1]
    return text.block_text(
        message,
        text.blue(text.l_(f.function, f.lineno, f.filename)), '',
        text.tbl(*(
            (f.function, f.lineno, f.filename, f'>>> {f.code_context[0].strip()}')
            for f in stack[offset+2:][:limit]
            if not fn or fn in f.filename
        )), ch=text.yellow('*')
    )

def print_stack(message=None, *a, offset=0, **kw):
    '''Print out the current stack trace.'''
    text.printasone(block_stack(message, *a, offset=offset + 1, **kw), flush=True)


def short_stack(match=None, file=False, sep=' << ', n=None):
    '''Print out a compressed view of the stack.'''
    return sep.join(
        (f'{f.function} ({os.path.basename(f.filename)}:{f.lineno})'
         if file else f.function)
        for f in inspect.stack()[1:][:n]
        if not match or match in f.filename
    )

def funcstr(func, *a, **kw):
    return '{}({})'.format(
        func.__name__, ', '.join(
            ['{!r}'.format(v) for v in a] +
            ['{}={!r}'.format(k, v) for k, v in kw.items()]
        ))

def excline(e):
    return '{}: {}'.format(type(e).__name__, e)

def print_excline(e):
    return print(excline)



from .meta import Meta
from . import statistics
from .stopwatch import Stopwatch
from .text import *
from .debug import *
from .misc import *
from . import logging
from . import shell
from . import func
from . import iters
from . import handlers
# from . import remote
from . import config
# from .background_process import *
from .writer import *



import time
import cv2
import reip


class Video(reip.Block):
    cap = None
    def __init__(self, index=None, file=None, fps=30, size=None, **kw):
        self.index = index
        self.size = size
        self.fps = fps
        super().__init__(n_inputs=None, **kw)

    def init(self):
        self.cap = cv2.VideoCapture(self.index)
        msg = ''
        if self.size:
            h, w = self.size
            wset = self.cap.set(cv2.CAP_PROP_FRAME_WIDTH, int(w))
            hset = self.cap.set(cv2.CAP_PROP_FRAME_HEIGHT, int(h))
            msg = f'  (desired window size: {h}x{w})'
        h = self.cap.get(cv2.CAP_PROP_FRAME_HEIGHT)
        w = self.cap.get(cv2.CAP_PROP_FRAME_WIDTH)
        self.log.debug(f'window size: {w}x{h}' + msg)

    def process(self, x=None, meta=None):
        if not self.cap.isOpened():
            print('closing', self.cap.isOpened())
            return reip.CLOSE
        ret, frame = self.cap.read()
        if not ret:
            raise RuntimeError(f'Video frame grab failed for {self}')
        return [frame[:, ::-1]], {
            'time': time.time(),
            'dim': frame.shape,
            'fps': self.fps
        }

    def finish(self):
        self.cap.release()


class _VideoWriter(reip.util.CycledWriter):
    # CODECS: XVID, MJPG, MPEG, MP4V
    def __init__(self, filename, duration=600,
                 codec='X264', **kw):
        self.duration = duration
        self.codec = codec
        super().__init__(filename)

    def _new_writer(self, fname, meta):
        if self.duration:
            self.file_length = int(meta['fps'] * self.duration)
        print(meta['dim'])
        return cv2.VideoWriter(
            fname, cv2.VideoWriter_fourcc(*self.codec),
            meta['fps'], meta['dim'][:2][::-1])

    def _close_writer(self):
        if self._writer is not None:
            self._writer.release()


class VideoWriter(reip.Block):
    def __init__(self, filename, **kw):
        self.filename = filename
        super().__init__(extra_kw=True, **kw)

    def init(self):
        self.writer = _VideoWriter(self.filename, **self.extra_kw)

    def process(self, X, meta):
        writer = self.writer.get(meta)
        writer.write(X)
        self.writer.increment()
        return self.writer.output_closed_file()

    def finish(self):
        self.writer.close()


# class VideoShow(reip.Block):
#     '''XXX: Doesn't work in multithreaded environments. use stream_imshow'''
#     def __init__(self, window_name=None, quit_key='q', **kw):
#         super().__init__(n_outputs=0, **kw)
#         self.window_name = window_name or self.name
#         self.quit_keys = {ord(k) for k in quit_key}
#
#     def process(self, X, meta):
#         cv2.imshow(self.window_name, X)
#         if cv2.waitKey(25) & 0xFF in self.quit_keys:
#             return reip.CLOSE
#         return [], {}
#
#     def finish(self):
#         cv2.destroyWindow(self.window_name)


def stream_imshow(stream, *names):
    try:
        for name in names:
            cv2.namedWindow(name)
        with stream:
            for xs, meta in stream:
                for name, x in zip(names, xs):
                    if x is not None:
                        cv2.imshow(name, x)
                if cv2.waitKey(25) & 0xFF == ord('q'):
                    return
                yield
    finally:
        for name in names:
            cv2.destroyWindow(name)



'''

# func_as_block : these are all equivalent

@reip.helpers.asblock(single_output=True, meta=False)
def basic_process_func(x, meta, n=2):
    return x * n

@reip.helpers.asblock(single_output=True)
def basic_process_func(x, meta, n=2):
    return x * n, {}

@reip.helpers.asblock(meta=False)
def basic_process_func(x, meta, n=2):
    return [x * n]

@reip.helpers.asblock()
def basic_process_func(x, meta, n=2):
    return [x * n], {}


# context_as_block (same variations apply for inner function)

@reip.helpers.asblock(single_output=True, meta=False, context=True)
def basic_process_func(n=2):
    def inner(x, meta):
        return x * n
    try:
        print('initialize')
        yield process
    finally:
        print('finish')
'''
import sys
import reip
import functools
import warnings
from contextlib import contextmanager



def asout(*values, meta=None):
    return list(values), ({} if meta is None else meta)



def asblock(__func__=None, single_output=False, n_inputs=None, meta=True, context=False, self=False, name=None, **kw):
    def asblock_converter(func):
        return type(name or func.__name__, (_BlockHelper,), {
            '__init__': reip.util.partial(
                _BlockHelper.__init__, _func=func, _single_output=single_output, _meta=meta,
                _context=context, _self=self, n_inputs=n_inputs, **kw)
        })
    return asblock_converter(__func__) if callable(__func__) else asblock_converter

def asbasic(*a, meta=False, **kw):
    '''Create a block using a process function that takes a single input and returns
    a single output.

    Example:
    >>> TimesN = reip.helpers.asbasic(lambda x, n=2: x * n)  # creates a block
    >>> B.Increment(5).to(TimesN(6)).to(B.Debug())  # 0 6 12 18 24
    '''
    return asblock(*a, single_output=True, meta=meta, **kw)

def asmulti(*a, n_inputs=None, meta=False, **kw):
    '''Create a block using a process function that takes multiple inputs and returns
    a multiple inputs.

    Example:
    >>> Sum = reip.helpers.asmulti((lambda *xs: [sum(xs)]), n_inputs=None)  # creates a block
    >>> B.Increment(5).to(Sum(6)).to(B.Debug())  # 0 6 12 18 24
    >>> out = Sum()(B.Increment(12), B.Increment(10, 22))
    '''
    return asblock(*a, single_output=False, meta=meta, n_inputs=n_inputs, **kw)

def asbasicmeta(*a, **kw):
    return asblock(*a, single_output=True, meta=True, **kw)

def asmultimeta(*a, **kw):
    return asblock(*a, single_output=False, meta=True, **kw)

def ascontext(*a, self=True, meta=True, single_output=False, **kw):
    return asblock(*a, context=True, self=self, meta=meta, single_output=single_output, **kw)

def asbasiccontext(*a, self=True, meta=False, single_output=True, **kw):
    return asblock(*a, context=True, self=self, meta=meta, single_output=single_output, **kw)

# internals


def wrap_process(single_output=False, meta=True):
    '''Handle multiple output formats. Convert them to a uniform format.'''
    def outer(func):
        if single_output and meta:  # return x, meta => [x], meta
            def inner(*a, **kw):
                x, meta = func(*a, **kw)
                return [x], meta
            return functools.wraps(func)(inner)
        if single_output and not meta:  # return x => [x], {}
            return functools.wraps(func)(lambda *a, meta, **kw: ([func(*a, **kw)], {}))
        if not single_output and not meta:  # return xs => xs, {}
            return functools.wraps(func)(lambda *a, meta, **kw: (func(*a, **kw), {}))
        return func  # return xs, meta
    return outer

def _wrap_context(func, context=False):
    '''Wrap a process function and convert it to a context manager.'''
    @functools.wraps(func)
    def inner(*a, **kw):
        yield reip.util.partial(func, *a, **kw)
    return contextmanager(func if context else inner)


class _BlockHelper(reip.Block):
    def __init__(self, *a, _func, _single_output=False, _context=False, _meta=True, _self=False, **kw):
        self.func = _func
        self.__single_output = _single_output
        self.__context = _context
        self.__meta = _meta
        self.extra_posargs = (self,)+a if _self else a
        super().__init__(extra_kw=True, **kw)

    _ctx = None
    def init(self):
        # ctx represents a context manager that does init and cleanup
        func = _wrap_context(self.func, self.__context)
        self._ctx = func(*self.extra_posargs, **self.extra_kw)
        # context should yield the process function.
        try:
            process = self._ctx.__enter__()
        except StopIteration as e:
            raise RuntimeError('Block context function did not yield.') from e
        self._process = wrap_process(single_output=self.__single_output, meta=self.__meta)(process)

    def process(self, *xs, meta):
        return self._process(*xs, meta=meta)

    def finish(self):
        self._ctx.__exit__(*sys.exc_info())



import os
import sys
import time
import json
import numpy as np
from random import choice
import requests
import reip

# import os
# import time
import glob
import traceback

ALPHABETICALLY = lambda f: f.split('/')[-1]

def get_upload_files(*patterns, interval=12, empty_interval=1, sort=ALPHABETICALLY, min_mtime=3):
    '''Infinitely return files from a directory, sorted by date.
    '''
    while True:
        try:
            # get file statistics
            files = sorted((
                f for pattern in patterns
                for f in glob.iglob(pattern)
            ), key=sort, reverse=True)

            ti = time.time()
            i = 0
            while i < len(files):
                # file list expires after certain time
                if time.time() - ti > (interval if files else empty_interval):
                    break

                fname = files[i]
                # file is too young
                if min_mtime and not check_mtime(fname, min_mtime):
                    continue
                # good to go, next
                yield fname
                i += 1
        except Exception:  # some unexpected exception - assume it's a weird thing and continue on
            traceback.print_exc()

def check_mtime(fname, min_mtime):
    if min_mtime:
        try:
            mtime = os.path.getmtime(fname)
            if time.time() - mtime >= min_mtime:
                return False
        except FileNotFoundError:  # file got deleted by something, just pop and ignore
            return False
    return True



def checkfile(f):
    return f if f and os.path.isfile(f) else None


class UploadError(requests.exceptions.RequestException):
    pass


class BaseUpload(reip.Block):
    def __init__(self, endpoint, servers=None, method='post', data=None,
                 cacert=None, client_cert=None,
                 client_key=None, client_pass=None, crlfile=None,
                 timeout=60, verify=True, n_tries=5, retry_sleep=15,
                 sess=None, fake=False, **kw):
        """Data Uploader.

        Args:
            cacert (str, optional): path for the CA certificate
            client_cert (str, optional): path for client's SSL certificate
            client_key (str, optional): path for client's SSL key
            client_pass (str, optional): Password for client's SSL cert
            crlfile (str, optional): Certificate Revocation List file for checking
                if the certificate has been revoked by CA
            timeout (int): seconds to wait before timing out
        """
        self.endpoint = endpoint
        self.servers = reip.util.as_list(servers)
        self.method = method
        self.data = data
        self.verify = verify
        self.n_tries = n_tries
        self.retry_sleep = retry_sleep
        self.timeout = timeout
        self.cacert = checkfile(cacert)
        self.client_cert = checkfile(client_cert)
        self.client_key = checkfile(client_key)
        self.client_pass = client_pass or None
        self.crlfile = checkfile(crlfile)
        self.fake = fake
        self._given_sess = sess
        super().__init__(**kw)

    sess = None
    def init(self):
        self.sess = self._given_sess or requests.Session()
        self.sess.protocol = 'http'
        if self.verify:
            self.sess.protocol = 'https'
            self.sess.verify = self.cacert or True
            if self.client_cert and self.client_key:
                self.sess.cert = (self.client_cert, self.client_key)

    def process(self, *files, meta):
        pass

    def finish(self):
        self.sess = None

    def get_url(self):
        # construct url
        return '{}://{}{}'.format(
            self.sess.protocol, choice(self.servers),
            self.endpoint) if self.servers else self.endpoint

    def request(self, **kw):
        # setup
        url = self.get_url()
        datastr = self.data_str(**kw)

        if self.fake:
            return requests.Response(url), 0, 0
        #print('request', kw)
        # retry request until it succeeds
        for i in reip.util.iters.loop():
            try:
                response = self.sess.request(self.method, url, **kw, timeout=self.timeout)
                # check for non-200 error code
                if not response.ok:
                    raise UploadError('Error when uploading {} to {}.\n{}'.format(
                        datastr, url, response.text))
            except requests.exceptions.RequestException as e:
                self.log.error(reip.util.excline(e))
                # exit if we've failed too many times
                if i >= self.n_tries-1:
                    e.response = response
                    self.on_failure(response, url, **kw)
                    raise e

                time.sleep(self.retry_sleep)
            else:
                # return response info
                secs = response.elapsed.total_seconds()
                speed = self.calc_size(**kw) / secs / 1024.0
                self.log.debug('{} uploaded to {} at {:.1f} Kb/s in {:.1f}s'.format(
                    datastr, url, speed, secs))

                self.on_success(response, url, **kw)
                return response, secs, speed

    def calc_size(self, **kw):
        return os.path.getsize(kw)

    def data_str(self, **kw):
        return ''

    def on_success(self, resp, url, **kw):
        pass

    def on_failure(self, resp, url, **kw):
        pass


class _AbstractUploadFile(BaseUpload):
    def process(self, files, meta=None, data=None):
        try:
            response, secs, speed = self.request(
                files=self.as_file_dict(files),
                data=data or reip.util.resolve_call(self.data, files, meta=meta))
            return [response], {'upload_time': secs, 'upload_kbs': speed, 'status_code': response.status_code}
        except Exception as e:
            return

    def as_file_dict(self, files):
        fileobjs = {name: self.open_file(fname) for name, fname in files.items()}
        return {k: (os.path.basename(fn), f) for k, (fn, f) in fileobjs.items()}

    def data_str(self, files, **kw):
        return ', '.join(os.path.join(k, fn) for k, (fn, f) in files.items())

    def open_file(self, fname):
        return fname, open(fname, 'rb')

    def calc_size(self, files, **kw):
        return sum(os.path.getsize(f) for f in files.values())


class UploadFile(_AbstractUploadFile):
    def __init__(self, *a, names=None, remove_on_success=True, **kw):
        self.names = reip.util.as_list(names or ())
        self.remove_on_success = remove_on_success
        super().__init__(*a, **kw)

    def process(self, *fs, meta=None):
        return super().process({
                reip.util.resolve_call(self.names[i], fname)
                if i < len(self.names) else
                os.path.basename(os.path.dirname(fname))
                or os.path.basename(fname): fname
                for i, fname in enumerate(fs)
            }, data=reip.util.resolve_call(self.data, *fs, meta=meta))

    def on_success(self, resp, url, files, **kw):
        if self.remove_on_success: # delete file
            for f in files.values():
                os.remove(f)



class UploadJSONAsFile(_AbstractUploadFile):
    FILENAME = 'status.json'
    def process(self, *data, meta=None):
        status = {k: v for d in data for k, v in d.items()}
        data = reip.util.resolve_call(self.data, status, meta=meta)
        return super().process({'file': status}, meta=meta, data=data)

    def open_file(self, status):
        return self.FILENAME, json.dumps(status, cls=NumpyEncoder)

    def calc_size(self, files):
        return sum(sys.getsizeof(x) for x in files.values())




class UploadJSON(BaseUpload):
    def process(self, *data, meta=None):
        status = {k: v for d in data for k, v in d.items()}
        try:
            data = reip.util.resolve_call(self.data, status, meta=meta) or {}
            response, secs, speed = self.request(json=dict(data, **status))
            return [response], {'upload_time': secs, 'upload_kbs': speed, 'status_code': response.status_code}
        except Exception as e:
            return

    def calc_size(self, data=None, json=None, **kw):
        return sys.getsizeof(data)








class UploadFilesFromDirectory(UploadFile):
    def __init__(self, endpoint, *patterns, n_inputs=0, **kw):
        self.patterns = patterns
        super().__init__(endpoint, n_inputs=n_inputs, **kw)

    def init(self):
        self.iter_files = get_upload_files(*self.patterns)

    def process(self, meta):
        fname = next(self.iter_files)
        return super().process(fname, meta=meta)





class NumpyEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        if isinstance(obj, np.generic):
            return obj.item()
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        return super().default(obj)




import os
import cv2
import reip
import tflit
import numpy as np

import reip


class Posenet(reip.Block):
    '''Pose Keypoint Detection.
    
    Outputs:
        keypoints (np.ndarray[n, (x, y, z)]): coordinates of the keypoints and flags for
            those that have low probability.
        drawn_image (np.ndarray[h, w, c]): The input image with keypoints drawn on.
    '''
    def __init__(self, draw=None, **kw):
        self._should_draw = draw
        super().__init__(n_outputs=2, **kw)

    def init(self):
        self.model = PosenetModel()
        self.should_draw = bool(self.sinks[1].readers) if self._should_draw is None else self._should_draw
        self.log.debug(f'will draw: {self.should_draw}')


    def process(self, frame, meta):
        frame = np.array(frame)
        kps = self.model.predict_batch(frame[None])
        return [kps, self.model.draw(frame, kps) if self.should_draw else None], meta


class PosenetModel(tflit.Model):
    '''Posenet - pose estimation
    
    Source: https://www.tensorflow.org/lite/examples/pose_estimation/overview
    '''
    MODEL_FILE = "posenet_mobilenet_v1_100_257x257_multi_kpt_stripped.tflite"
    multi_input = False
    def __init__(self, threshold=0.3):
        super().__init__(reip.package_file('models', self.MODEL_FILE))
        self.summary()
        self.size = self.input_shape[1:-1]
        self.threshold = threshold

    def predict_batch(self, image, *a, **kw):
        small = cv2.resize(image[0], self.size)
        x = (np.float32(small) - 127.5) / 127.5
        heat, offset = super().predict_batch(x[None], *a, **kw)[:2]
        kps = parse_output(heat[0], offset[0], self.threshold)
        return kps

    def draw(self, X, kps, **kw):
        return draw_kps(X, kps, size=self.size)



def parse_output(heatmap, offset, threshold=0.3):
    '''Parse Posenet outputs.

    Args:
        heatmap (np.array[d=3]): heatmaps for an image. First output of model.
        offset (np.array[d=3]): offset vectors for an image. Second output of model.
        threshold (float): probability threshold for the keypoints.

    Returns:
        kps (np.array[n x (x,y,z)]): coordinates of the keypoints and flags for
            those that have low probability.
    '''

    joint_num = heatmap.shape[-1]
    pose_kps = np.zeros((joint_num, 3), np.uint32)

    for i in range(joint_num):
        joint_heatmap = heatmap[...,i]
        max_prob = np.max(joint_heatmap)
        i_max = np.squeeze(np.argwhere(joint_heatmap == max_prob))
        remap_pos = np.array(i_max/8*257, dtype=np.int32)
        pose_kps[i, 0] = int(remap_pos[0] + offset[i_max[0], i_max[1], i])
        pose_kps[i, 1] = int(remap_pos[1] + offset[i_max[0], i_max[1], i+joint_num])
        if max_prob > threshold and pose_kps[i, 0] < 257 and pose_kps[i, 1] < 257:
            pose_kps[i, 2] = 1
    return pose_kps

def draw_kps(img, kps, limbs=True, face=True, size=None, thickness=3):
    if size:
        ratio = np.array(tuple(x/y for x, y in zip(img.shape, size)) + (1,))
        kps = (kps * ratio).astype(kps.dtype)
    for i in range(0 if face else 5, len(kps)):
        if kps[i, 2]:
            cv2.circle(img, (kps[i, 1], kps[i, 0]), 2, color=(255, 0, 255), thickness=thickness)

    if limbs:
        for part in body_parts:
            kp1, kp2 = kps[part[0]], kps[part[1]]
            if kp1[2] and kp2[2]:
                cv2.line(
                    img, (kp1[1], kp1[0]), (kp2[1], kp2[0]),
                    color=(255, 255, 255), lineType=cv2.LINE_AA,
                    thickness=thickness)
    return img


body_parts = [(5,6),(5,7),(6,8),(7,9),(8,10),(11,12),(5,11),(6,12),(11,13),(12,14),(13,15),(14,16)]


Pose = Posenet


import reip
import numpy as np
import librosa


def dbmean(X):
    return 10*np.log10(np.mean(10**(X/10)))


def as_power_of_2(x):
    return int(2 ** np.floor(np.log2(x)))

class SPL(reip.Block):
    weightings = None
    hop_duration = None
    n_fft = None
    def __init__(self, duration=1, weighting='zac',
                 calibration=0, **kw):
        super().__init__(extra_kw=True, **kw)
        self.duration = duration
        self.calibration = calibration
        self.weight_names = [w.upper() for w in list(weighting or 'Z')]
        # stft args
        self.extra_kw.setdefault('center', False)

    def _init(self, meta):
        # calculate the fft size using the sample rate and window duration
        self.n_fft = as_power_of_2(self.duration * meta['sr'])
        self.extra_kw.setdefault('hop_length', self.n_fft)
        self.hop_duration = self.extra_kw['hop_length'] / meta['sr']
        # get the frequency weights for each weighting function
        freqs = librosa.fft_frequencies(sr=meta['sr'], n_fft=self.n_fft)
        self.weightings = librosa.db_to_power(
            librosa.multi_frequency_weighting(freqs, kinds=self.weight_names))

    def process(self, data, meta):
        if self.weightings is None:  # initialize using sr from metadata
            self._init(meta)
        # select single channel (48000, 1) => (48000,)
        data = np.atleast_2d(data.T)[0, :]
        # get spectrogram (frequency, time) - (1025, 94)
        S = np.abs(librosa.core.stft(data, n_fft=self.n_fft, **self.extra_kw))
        # get weighted leq (3, 1025) x (1025, 94) => (94, 3)
        leq = self.calibration + librosa.power_to_db((
            self.weightings[..., None] * (S ** 2)).mean(-2)).T
        meta["hop_duration"] = self.hop_duration
        return [leq], meta
        # return [leq], {
        #     #'spl_weightings': self.weight_names,
        #     'hop_duration': self.hop_duration,
        # }


class Stft(reip.Block):
    def process(self, y, meta):
        S = librosa.amplitude_to_db(np.abs(librosa.stft(
            y[:, 0] if y.ndim > 1 else y,
            **self.extra_kw)))
        return [S], meta



import os
import reip
import matplotlib
matplotlib.use('agg')
import librosa.display
import matplotlib.pyplot as plt


class Specshow(reip.Block):
    def __init__(self, filename='{time}.png', figsize=(12, 6), cmap='magma', **kw):
        self.filename = filename
        self.figsize = figsize
        self.cmap = cmap
        self.kw = kw
        super().__init__()

    def process(self, X, meta):
        # get filename
        fname = self.filename.format(**meta)
        os.makedirs(os.path.dirname(fname), exist_ok=True)
        # plot
        plt.figure(figsize=self.figsize)
        librosa.display.specshow(X, cmap=self.cmap, **self.kw)
        plt.savefig(fname)
        return [fname], meta
